// Copyright 1992, 1999 by Jon Dart.  All Rights Reserved.

#ifndef _HASH_H
#define _HASH_H

#include "chess.h"
#include "stats.h"

extern const hash_t rep_codes[3];

class HashEntry;

class PositionInfo {
    friend class HashEntry;
public:

    enum ValueType { Invalid, NoHit, Valid, UpperBound, LowerBound};

    enum FLAGS { FORCED_FLAG = 0x80, IN_CHECK_FLAG = 0x40,
                 FULL_FLAG = 0x20, AGE_FLAG = 0x18 };

    enum { DEPTH_MASK = 0x3ff, PROC_MASK = 0xfc00 };

    enum { PROC_INC = 0x0400 };

    PositionInfo(int val = 0,int d = 0,
                  unsigned int flag = 0,Move best_move = NullMove) 
       :my_value(val),
        my_depth(d),
        my_flags((byte)flag),
        start(StartSquare(best_move)),
        dest(DestSquare(best_move)),
        promotion(PromoteTo(best_move)) {
    }

    ValueType type() const
    {
        return (ValueType)(my_flags & 0x7);
    }
    
    int value() const
    {
        return (int)my_value;
    }

    void setValue(int new_value) {
        my_value = (int16)new_value;
    }
    
    void copy(const HashEntry &source) {
      memcpy(&my_value,&((PositionInfo*)&source)->my_value,sizeof(PositionInfo));
    }

    int forced() const
    {
         return (my_flags & FORCED_FLAG) != 0;
    }
    
    int in_check() const
    {
        return (my_flags & IN_CHECK_FLAG) != 0;
    }

    int isFull() const
    {
        return (my_flags & FULL_FLAG) != 0;
    }

    int age() const
    {
        return (my_flags & AGE_FLAG) >> 3;
    }

    void setFull(int b) {
        if (b) my_flags |= (byte)FULL_FLAG;
        else my_flags &= (byte)~FULL_FLAG;
    }

    int depth() const {
        return (int)(my_depth & DEPTH_MASK);
    }

    byte flags() const {
        return my_flags;
    }

    Move best_move(const Board &b) const;

private:
    int16 my_value;
    int16 my_depth; 
    byte my_flags;
    signed char start, dest, promotion;
};    

class HashEntry : public PositionInfo {

    // this class represents one entry in the hash table.
    friend class Search;
    
    public:

    HashEntry( hash_t hashCode, const int depth,
                              byte age,
	           	      PositionInfo::ValueType hashflags,
                              int in_check,
			      int value,
 		              int full,
			      Move best_move)
   :  
      PositionInfo(value,
                    depth,
                    (IsForced(best_move)<<7) | (in_check<<6) | (age<<3) | (full<<5) | hashflags,
                    best_move),
      my_hashcode(hashCode)
   {}

   void update(const PositionInfo &info, int age) {
       memcpy(&my_value,&info,sizeof(PositionInfo));
       my_flags &= (uint16)~PositionInfo::AGE_FLAG;
       my_flags |= (age<<3);
    }
    
    hash_t hash_code() const  {
        return my_hashcode;
    }

    void update_best_move(const HashEntry &h) {
       if (h.start != InvalidSquare) {
         start = h.start;
         dest = h.dest;
         promotion = h.promotion;
       }
    }

    private:

    hash_t my_hashcode;
};

unsigned long hash_code(const HashEntry &p);

struct DualHashEntry {
   HashEntry replace;
   HashEntry depth;
};

// size is in kilobytes
extern void initHash(int size);

extern void resizeHash(int size);

extern void freeHash();

extern void clearHash();

// put info from the external permanent hash table into the 
// in-memory hash table
extern void loadLearnInfo();

extern PositionInfo::ValueType searchHash(const Board &b,
  hash_t hashCode,
  int alpha, int beta, int ply,
  int depth,
  PositionInfo &entry
);

extern void storeHash(const HashEntry &newPos,Statistics *stats = NULL);

#endif
