// Copyright 1992-2008 by Jon Dart. All Rights Reserved.
//
#ifndef _MOVE_GENERATOR_H
#define _MOVE_GENERATOR_H

#include "board.h"
#include "constant.h"

class SearchContext;
struct NodeInfo;

class MoveGenerator {
    public:
		
    enum Phase {HASH_MOVE_PHASE, WINNING_CAPTURE_PHASE,
      KILLER1_PHASE, KILLER2_PHASE, HISTORY_PHASE1,
      HISTORY_PHASE2,
      LOSERS_PHASE, LAST_PHASE };

    MoveGenerator( const Board &,
                    SearchContext *sc = NULL,
                    unsigned ply = 0, 
                    Move pvMove = NullMove, 
                    int trace = 0);
	
    // Generate the next move, in sorted order, NullMove if none left
    virtual Move nextMove() {
       if (index >= batch_count) {
          if ((batch_count = getBatch(batch,index)) == 0) 
             return NullMove;
       }
       return batch[index++];
    }

    virtual ~MoveGenerator() {
    }

    // Generate the next check evasion, NullMove if none left
    virtual Move nextEvasion();

    virtual Move nextMove(NodeInfo *node);
    virtual Move nextEvasion(NodeInfo *node);

   virtual int generateAllMoves(NodeInfo *);

    // Generate only non-capturing moves.
    int generateNonCaptures(Move *moves);

    // fill array "moves" with moves, starting at "index", returns
    // # of moves (0 if no more). This tries to generate moves in stages,
    // so it should be called repeatedly until the return value is zero.
    // 		
    // The moves returned are generally "pseudo-legal", i.e. they may
    // involve moves into check.  However, if the side to move is in
    // check, then all moves returned are strictly legal.		
    //
    int generateAllMoves(Move *moves,int repeatable);

    // Only generates captures, promotions, and check evasions.
    //
    int generateCaptures( Move *moves );

    // Return a list of the moves that may be used to escape
    // check.  Unlike the regular move generation routine, all
    // moves are checked for legality.
    //
    int generateEvasions(Move * moves);

    // Generate non-capturing checking moves
    int generateChecks(Move * moves);

    unsigned movesGenerated() const
    {
       return moves_generated;
    }

    Phase getPhase() const {
      return (Phase)phase;
    }

    int more() const {
      return phase<LAST_PHASE || index < batch_count;
    }
		
    static void sortMoves(Move moves[], int scores[], int n);

 protected:
		
    int getBatch(Move *&batch,int &index);
    int orderWinningCaptures (int captures, Move hashMove);

    int getEvasions();

    int generateEvasionsCaptures(Move * moves);
    int generateEvasionsNonCaptures(Move * moves);
    int generateEvasions(Move * moves,
			 const Bitboard &mask);

    const Board &board;
    SearchContext *context;
    int ply;
    int moves_generated;
    int phase, winners, losers_count,index,batch_count,forced;
    Move hashMove;
    int non_captures_count,history_count;
    Bitboard king_attacks; // for evasions
    int num_attacks; // for evasions
    Square source; // for evasions;
    Move *batch;
    Move losers[100];
    Move moves[Constants::MaxMoves];
    Move killer1,killer2,killer3,killer4;
    int master;
        
    inline void setMove( Square source, Square dest,
       PieceType promotion,
       Move *moves, unsigned &NumMoves);

};

struct MoveEntry {
   Move move;
   uint64 nodes;
};

class RootMoveGenerator : public MoveGenerator {
 public:
    RootMoveGenerator(const Board &board, 
                        SearchContext *context,
                        Move pvMove, 
		        int trace);

    // Generate the next move, in sorted order, NullMove if none left
    virtual Move nextMove() {
      ASSERT(index<=batch_count);
      if (index < batch_count)
        return moveList[index++].move;
      else
	return NullMove;
    }

    // Generate the next check evasion, NullMove if none left
    virtual Move nextEvasion() {
      return nextMove();
    }


    using MoveGenerator::nextMove;
    using MoveGenerator::nextEvasion;
    virtual int generateAllMoves(NodeInfo *);

    void reorder(Move pvMove,int depth);

    void exclude(Move *exclude, int num_exclude);

    void exclude(Move);

    int moveCount() const {
      return batch_count;
    }

    Move first() const {
      return moveList[0].move;
    }

    uint64 *getNodeCount(Move m) {
      for (int i=index-1;i>=0;i--) {
        // make sure we have the right move slot (when SMP enabled)
	if (MovesEqual(m,moveList[i].move)) {
	  return &moveList[i].nodes;
	}
      }
      ASSERT(0);
      return NULL;
    }

 private:
    int order();
    int score(const Move & move);

    MoveEntry moveList[Constants::MaxMoves];
};


#endif
