// Copyright 1994-2009 by Jon Dart.  All Rights Reserved.

#include "scoring.h"
#include "util.h"
#include "bhash.h"
#include "bitprobe.h"
extern "C" {
#include <stdlib.h>
};

#include <iomanip>

//#define PAWN_DEBUG
//#define EVAL_DEBUG
//#define LAZY_DEBUG

static const int LAZY_MARGIN = 7*PAWN_VALUE/2;

static Bitboard kingCoverW0[64];
static Bitboard kingCoverW1[64];
static Bitboard kingCoverW2[64];
static Bitboard kingCoverB0[64];
static Bitboard kingCoverB1[64];
static Bitboard kingCoverB2[64];
static Bitboard w_pawn_attacks[64], b_pawn_attacks[64];

static const int PAWN_HASH_SIZE = 8192;
static const int KING_COVER_HASH_SIZE = 16384;
static const int ENDGAME_HASH_SIZE = 32768;
static const int EVAL_CACHE_SIZE = 131072;

static const int CENTER_PAWN_BLOCK=-8;

static const int KING_COVER0 = 8;
static const int KING_COVER1 = 20;
static const int KING_COVER2 = 10;
static const int KING_OFF_BACK_RANK = -16;
static const int KING_NEAR_OPEN_FILE = -25;
static const int KING_NEAR_HALF_OPEN_FILE = -15;
static const int KING_PAWN_PROXIMITY[8] =
  {4, 4, 2, 1, 0, 0, 0, 0 };
static const int KING_ATTACK_MATERIAL_SCALE[32] = {
  0, 0, 0, 0, 0, 0, 0, 0, 
  2, 4, 7, 10, 13, 15, 17, 19,
  21, 22, 23, 25, 26, 27, 27, 28,
 28, 29, 29, 30, 30, 31, 32, 32};
static const int ATTACK_COUNT_SCALE[24] = {
  0, 10, 20, 30, 40, 50, 65, 80, 95, 110, 130, 150, 170,
  185, 200, 210, 220, 225, 230, 235, 240, 245, 250, 255
};
static const int PIN_MULTIPLIER = 15;
static const int PROXIMITY_MULTIPLIER = 4;
#define BOOST
static const int KING_ATTACK_MAX_BOOST = PAWN_VALUE;
static const int KING_ATTACK_BOOST_THRESHOLD = 33;

static const int OPPOSITION = 9;        // twice this for direct opposition

static const int PARTIAL_SCORE_BIAS = 20000;

struct EvalCacheEntry {
   hash_t score_key, move_key;
   int score;
   Move best;
};

static EvalCacheEntry *evalCache = NULL;

static const CACHE_ALIGN int KnightScores[64] =
{
    -18, -10, -10, -10, -10, -10, -10, -18,
    -8, 0, 4, 4, 4, 4, 0, -8,
    -8, 4, 8, 4, 4, 8, 4, -8,
    -8, 4, 4, 8, 8, 4, 4, -8,
    -8, 6, 6, 12, 12, 6, 6, -8,
    -6, 6, 10, 6, 6, 10, 6, -6,
    -8, 0, 4, 4, 4, 4, 0, -8,
    -8, 0, 0, 0, 0, 0, 0, -8
};

static const CACHE_ALIGN int BishopScores[64] =
{
    -12, -12, -12, -12, -12, -12, -12, -12,
    0, 8, 0, 6, 6, 0, 8, 0,
    0, 0, 6, 6, 6, 6, 0, 0,
    0, 0, 6, 6, 6, 6, 0, 0,
    0, 6, 6, 6, 6, 6, 6, 0,
    0, 0, 6, 6, 6, 6, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0
};

static const int OutpostScoresW[64] =
{
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 12, 16, 16, 12, 0, 0,
    0, 16, 16, 20, 20, 16, 16, 0,
    0, 12, 12, 12, 12, 12, 12, 0,
    0, 0, 0, 0, 0, 0, 0, 0
};

static const int OutpostScoresB[64] =
{
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 12, 12, 12, 12, 12, 12, 0,
    0, 16, 16, 20, 20, 16, 16, 0,
    0, 0, 12, 16, 16, 12, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0,
    0, 0, 0, 0, 0, 0, 0, 0
};

static const int KingEndgameScoresW[64] = {
-38, -34, -30, -29, -29, -30, -34, -38,
-29, -24, -21, -19, -19, -21, -24, -29,
-19, -14, -11, -10, -10, -11, -14, -19,
 -10, -5, -2, 0, 0, -2, -5, -10,
  0, 5, 6, 10, 10, 6, 5, 0,
  10, 14, 16, 19, 19, 16, 14, 10,
  14, 19, 21, 24, 24, 21, 19, 14,
  14, 19, 21, 24, 24, 21, 19, 14
};

static const int KingEndgameScoresB[64] = {
  14, 19, 21, 24, 24, 21, 19, 14,
  14, 19, 21, 24, 24, 21, 19, 14,
  10, 14, 16, 19, 19, 16, 14, 10,
  0, 5, 6, 10, 10, 6, 5, 0,
 -10, -5, -2, 0, 0, -2, -5, -10,
-19, -14, -11, -10, -10, -11, -14, -19,
-29, -24, -21, -19, -19, -21, -24, -29,
-38, -34, -30, -29, -29, -30, -34, -38};

static const int KBNKScores[64] =
{
    -40, -30, -15, -5, 5, 15, 30, 40,
    -30, -15, -5, 0, -1, 5, 15, 30,
    -15, -5, 0, 0, 0, -1, 5, 15,
    -5, -5, 0, -3, -3, 0, -1, 5,
    5, -1, 0, -3, -3, 0, 0, -5,
    15, 5, -1, 0, 0, 0, -5, -15,
    30, 15, 5, -1, 0, -5, -15, -30,
    40, 30, 15, 5, -5, -15, -30, -40
};

static const int BAD_BISHOP = -6;
static const int BISHOP_TRAPPED = -160;       // bishop trapped by pawn
static const int BISHOP_PAIR = 4*PAWN_VALUE/10;
static const int BISHOP_ENDGAME_BONUS = 10;
static const int BISHOP_PAIR_PAWN_ADJUST[9] =
{
  PAWN_VALUE/8, PAWN_VALUE/8, PAWN_VALUE/8, PAWN_VALUE/8,
  0, -PAWN_VALUE/8, -PAWN_VALUE/8, -PAWN_VALUE/8, -PAWN_VALUE/8
};
static const int RB_TRADE[8] =
{
  0, -PAWN_VALUE/8, 0, PAWN_VALUE/2, PAWN_VALUE/2,
  PAWN_VALUE/2, PAWN_VALUE/2, PAWN_VALUE/2
};
static const int RBN_TRADE[8] =
{
  0, PAWN_VALUE/2, PAWN_VALUE/4, PAWN_VALUE/4, PAWN_VALUE/4,
  PAWN_VALUE/4, PAWN_VALUE/4, PAWN_VALUE/4
};

static const int CASTLING_SCORES[6] =
{
    3, -12, -18, 16, 12, -24
};

// scores by rank
static const int PASSED_PAWN[8] =                 // score by rank
{0, 0, 10, 18, 30, 45, 90, 150};
static const int POTENTIAL_PASSER[8] =            // score by rank
{0, 0, 10/2, 18/2, 30/2, 45/2, 90/2, 150/2 };
static const int PASSED_PAWN_BLOCK[8] =           // score by rank
{0, 0, 10/3, 18/3, 30/3, 45/4, 90/4, 150/5 };
static const int OUTSIDE_PP[16] =
{
  45, 36, 24, 21, 18, 18, 15, 15,
  15, 15, 12, 12, 12, 12, 12, 12
};
static const int CONNECTED_PASSERS[8] =
{0, 0, 10, 12, 15, 25, 40, 75 };
static const int ADJACENT_PASSERS = 12;
static const int DOUBLED_PAWNS = -16;
static const int TRIPLED_PAWNS = -16;             // in addition to double penalty
static const int ISOLATED_PAWN[8] = {0,-8,-16,-32,-64,-64,-64,-64,};
static const int WEAK = -6;
static const int WEAK2 = -6;
static const int WEAK_ON_OPEN_FILE = -10;
static const int SPACE=2;
static const int PAWN_CENTER_SCORE=6;
static const int PAWN_CENTER_SCORE2=5;

static const int BISHOP_MOBILITY = 3;

static const int KNIGHT_MOBILITY[9]={-12,-6,-2,0,2,4,6,8,10};

static const int ROOK_ON_7TH_RANK = 40;
// rook on open file or file with weak pawn:
static const int ROOK_ON_OPEN_FILE = 15;
static const int DOUBLED_ROOKS = 15;
static const int ROOK_MOBILITY[16] = {-15,-10,-6,-3,0,3,5,6,
                                      7,8,9,10,10,10,10,10};
static const int QUEEN_OUT = -6;
static const int ROOK_BEHIND_PP = 20;
static const int ROOK_AHEAD_OF_PP = -10;
static const int SIDE_PROTECTED_PAWN = -10;

// endgame terms
static const int UNCATCHABLE_PAWN = 12;
static const int BITBASE_WIN=500;
static const int KING_AHEAD_OF_PAWN = 3;
static const int DISTANCE_FROM_PAWN = -4;
static const int SUPPORTED_PASSER6 = 38;
static const int SUPPORTED_PASSER7 = 76;

static Bitboard backwardW[64],backwardB[64];
CACHE_ALIGN Bitboard passedW[64],passedB[64];       // not static because needed by search module
static Bitboard weakW[64],weakB[64];
static Bitboard outpostW[64],outpostB[64];
static Bitboard connected_passers[64][2];
static Bitboard adjacent_passers[64];
static Bitboard black_squares,white_squares;
static Bitboard rook_pawn_mask(Attacks::file_mask[0] | Attacks::file_mask[7]);
static Bitboard abcd_mask,efgh_mask;
static Bitboard left_side_mask[8], right_side_mask[8];
static Bitboard center;

int distances[64][64];
static int opposition[64][64];

// Don't call it "distance" - conflicts with MS template library
static inline int distance1(int x, int y)
{
    return distances[x][y];
}


static inline int OnFile(const Bitboard &b, int file)
{
    return TEST_MASK(b,Attacks::file_mask[file-1]);
}


static inline int FileOpen(const Board &board, int file, ColorType side)
{
  return !TEST_MASK(board.pawn_bits[side],Attacks::file_mask[file-1]);
}


static inline int FileOpen(const Board &board, int file)
{
  return !TEST_MASK((board.pawn_bits[White] | board.pawn_bits[Black]),Attacks::file_mask[file-1]);
}


// Return all squares attacked by a pawn of color "side" at square "sq".
static FORCEINLINE Bitboard pawn_attacks( const Board &board, Square sq, ColorType side)
{
    return Bitboard(Attacks::pawn_attacks[sq][side] & board.pawn_bits[side]);
}


static void initBitboards()
{
    int i, r;

    for (i = 0; i < 64; i++) {
        if (SquareColor(i) == White)
            white_squares.set(i);
        else
            black_squares.set(i);
        kingCoverW0[i] = kingCoverB0[i] = Bitboard(0);
        if (File(i) != 8) {
            kingCoverW0[i].set(i+1);
            kingCoverB0[i].set(i+1);
        }
        if (File(i) != 1) {
            kingCoverW0[i].set(i-1);
            kingCoverB0[i].set(i-1);
        }
        if (Rank(i,White) != 8) {
            kingCoverW1[i].set(i+8);
            if (File(i) != 8) {
                kingCoverW1[i].set(i+9);
            }
            if (File(i) != 1) {
                kingCoverW1[i].set(i+7);
            }
        }
        kingCoverW1[H1].set(F2);
        kingCoverW1[A1].set(C2);
        kingCoverB1[H8].set(F7);
        kingCoverB1[A8].set(C7);
        if (Rank(i,White)<=2)
            kingCoverW2[i].set(i+16);
        if (Rank(i,White) < 7) {
            if (File(i) != 8) {
                ASSERT(i+9+8<64);
                kingCoverW2[i].set(i+9+8);
            }
            if (File(i) != 1) {
                ASSERT(i+7+8<64);
                kingCoverW2[i].set(i+7+8);
            }
        }
        if (Rank(i,Black) < 7) {
            ASSERT(i-9>=0);
            kingCoverB1[i].set(i-8);
            if (File(i) != 8) {
                kingCoverB1[i].set(i-7);
            }
            if (File(i) != 1) {
                kingCoverB1[i].set(i-9);
            }
        }
        if (Rank(i,Black)<=2)
            kingCoverB2[i].set(i-16);
        if (Rank(i,Black) < 7) {
            if (File(i) != 1) {
                ASSERT(i-9-8>=0);
                kingCoverB2[i].set(i-9-8);
            }
            if (File(i) != 8) {
                ASSERT(i-7-8>=0);
                kingCoverB2[i].set(i-7-8);
            }
        }
        kingCoverW2[H1].set(F3);
        kingCoverW2[H2].set(F3);
        kingCoverW2[A1].set(C3);
        kingCoverW2[A2].set(C3);
        kingCoverB2[H8].set(F6);
        kingCoverB2[H7].set(F6);
        kingCoverB2[A8].set(C6);
        kingCoverB2[A7].set(C6);

        int file = File(i);
        int rankw = WhiteRank(i);
        int fmin = Util::Max(1,file-2);
        int fmax = Util::Min(8,file+2);
        for (int f = fmin; f <= fmax; f++) {
            int r;
            if (Util::Abs(f-file)<2) {
                for (r = rankw; r <= Util::Min(rankw+3,8); r++) {
                    w_pawn_attacks[i].set(MakeSquare(f,r,White));
                }
                for (r = rankw; r >= Util::Max(rankw-3,1); r--) {
                    b_pawn_attacks[i].set(MakeSquare(f,r,White));
                }
            }
            else {
                for (r = rankw; r <= Util::Min(rankw+3,8); r++) {
                    w_pawn_attacks[i].set(MakeSquare(f,r,White));
                }
                for (r = rankw; r >= Util::Max(rankw-3,1); r--) {
                    b_pawn_attacks[i].set(MakeSquare(f,r,White));
                }
            }
        }

        for (int j = 0; j < 64; j++) {
            int file_distance, rank_distance;
            Square kp = i;
            Square oppkp = j;
            file_distance=Util::Abs(File(kp)-File(oppkp));
            rank_distance=Util::Abs(Rank(kp,Black)-Rank(oppkp,Black));
            if (file_distance == 0 && rank_distance == 2) {
                opposition[i][j] = 2;             // direct opposition
            }
            else if (file_distance == 0 && rank_distance % 2 == 0)
                opposition[i][j] = 1;             // remote opposition
            else if ((file_distance == 2) && (rank_distance == 2))
                opposition[i][j] = 1;
            else
                opposition[i][j] = 0;
            if (file_distance > rank_distance)
                distances[i][j] = file_distance;
            else
                distances[i][j] = rank_distance;
        }

        int rank = Rank(i,White);
        for (r = rank; r > 1; r--) {
            Square sq2 = MakeSquare(file,r,White);
            if (file != 8) {
                sq2 = MakeSquare(file+1,r,White);
                backwardW[i].set((int)sq2);
            }
            if (file != 1) {
                sq2 = MakeSquare(file-1,r,White);
                backwardW[i].set((int)sq2);
            }
        }
        weakW[i] = backwardW[i];
        if (file != 8) {
            weakW[i].clear(MakeSquare(file+1,rank,White));
        }
        if (file != 1)
            weakW[i].clear(MakeSquare(file-1,rank,White));
        for (r = rank + 1; r < 8; r++) {
            Square sq2 = MakeSquare(file,r,White);
            passedW[i].set(sq2);
            if (file != 8) {
                sq2 = MakeSquare(file+1,r,White);
                passedW[i].set(sq2);
                outpostW[i].set(sq2);
            }
            if (file != 1) {
                sq2 = MakeSquare(file-1,r,White);
                passedW[i].set(sq2);
                outpostW[i].set(sq2);
            }
        }
        rank = Rank(i,Black);
        for (r = rank; r > 1; r--) {
            Square sq2 = MakeSquare(file,r,Black);
            if (file != 8) {
                sq2 = MakeSquare(file+1,r,Black);
                backwardB[i].set(sq2);
            }
            if (file != 1) {
                sq2 = MakeSquare(file-1,r,Black);
                backwardB[i].set(sq2);
            }
        }
        weakB[i] = backwardB[i];
        if (file != 8) {
            weakB[i].clear(MakeSquare(file+1,rank,Black));
            connected_passers[i][White].set(i+1);
            connected_passers[i][Black].set(i+1);
            adjacent_passers[i] |= Attacks::file_mask[file];
        }
        if (file != 1) {
            weakB[i].clear(MakeSquare(file-1,rank,Black));
            connected_passers[i][White].set(i-1);
            connected_passers[i][Black].set(i-1);
            adjacent_passers[i] |= Attacks::file_mask[file-2];
        }
        connected_passers[i][Black] |= Attacks::pawn_attacks[i][Black];
        connected_passers[i][White] |= Attacks::pawn_attacks[i][White];
        for (r = rank + 1; r < 8; r++) {
            Square sq2 = MakeSquare(file,r,Black);
            passedB[i].set((int)sq2);
            if (file != 8) {
                sq2 = MakeSquare(file+1,r,Black);
                passedB[i].set(sq2);
                outpostB[i].set(sq2);
            }
            if (file != 1) {
                sq2 = MakeSquare(file-1,r,Black);
                passedB[i].set(sq2);
                outpostB[i].set(sq2);
            }
        }
        if (file<EFILE) abcd_mask.set(i);
        else efgh_mask.set(i);
    }
    int x;
    for (x=1; x <8; x++) {
        for (int y=x-1;y>=0;y--)
            left_side_mask[x] |= Attacks::file_mask[y];
    }
    for (x=6; x >=0; x--) {
        for (int y=x+1;y<8;y++)
            right_side_mask[x] |= Attacks::file_mask[y];
    }
    center.set(D4);
    center.set(D5);
    center.set(E4);
    center.set(E5);
}


void Scoring::init()
{
    initBitboards();
    if (evalCache == NULL) {
       evalCache = (EvalCacheEntry*)malloc(EVAL_CACHE_SIZE*sizeof(EvalCacheEntry));    
       for (int i = 0; i < EVAL_CACHE_SIZE; i++) {
         evalCache[i].score_key = evalCache[i].move_key = (hash_t)0;
         evalCache[i].score = INVALID_SCORE;
         evalCache[i].best = NullMove;
       }
    }
}

void Scoring::cleanup() {
    delete evalCache;
}

Scoring::Scoring()
{
    ALIGNED_MALLOC(pawnHashTable,PawnHashEntry, PAWN_HASH_SIZE*sizeof(PawnHashEntry),128);
    memset(pawnHashTable,'\0',PAWN_HASH_SIZE*sizeof(PawnHashEntry));

    ALIGNED_MALLOC(kingCoverHashTable,KingCoverHashEntry,KING_COVER_HASH_SIZE*sizeof(KingCoverHashEntry),128);
    memset(kingCoverHashTable,'\0',KING_COVER_HASH_SIZE*sizeof(KingCoverHashEntry));
    endgameHashTable = (EndgameHashEntry*)malloc(ENDGAME_HASH_SIZE*sizeof(EndgameHashEntry));
    memset(endgameHashTable,'\0',ENDGAME_HASH_SIZE*sizeof(EndgameHashEntry));
}



Scoring::~Scoring()
{
    ALIGNED_FREE(pawnHashTable);
    ALIGNED_FREE(kingCoverHashTable);
    free(endgameHashTable);
}


static int adjustMaterialScore(const Board &board, ColorType side)
{
    const Material &ourmat = board.getMaterial(side);
    const Material &oppmat = board.getMaterial(OppositeColor(side));
    int score = 0;
    if (ourmat.bishopCount() >=2 && oppmat.bishopCount() < 2) {
        int pawns = ourmat.pawnCount() + oppmat.pawnCount();
        score += (BISHOP_PAIR + BISHOP_PAIR_PAWN_ADJUST[pawns/2]);
#ifdef EVAL_DEBUG
        cout << "bishop pair (" << ColorImage(side) << ") : " <<
            (BISHOP_PAIR + BISHOP_PAIR_PAWN_ADJUST[pawns/2]) << endl;
#endif
    }
    // Evaluate uneven trades. Penalties/bonuses loosely based on
    // Larry Kaufman article.
    int trade_score = 0;
    if (Util::Abs(ourmat.value()-oppmat.value())<=5*PAWN_VALUE) {
        int oppMinors = oppmat.bishopCount() + oppmat.knightCount();
        int ourMinors = ourmat.bishopCount() + ourmat.knightCount();
        if (ourMinors > oppMinors) {
            int ourQ = ourmat.queenCount();
            int oppQ = oppmat.queenCount();
            int ourR = ourmat.rookCount();
            int oppR = oppmat.rookCount();
            if (ourQ == oppQ && ourR == oppR) {
                // knight or bishop vs. pawns
                if (ourMinors == oppMinors+1) {
                    if (ourmat.infobits() == Material::KB ||
                       ourmat.infobits() == Material::KN) {
                        // we don't have mating material (no pawns)
                        trade_score-=PAWN_VALUE;
                    }
                    else                          // small bonus for piece
                        trade_score+=PAWN_VALUE/4;
#ifdef EVAL_DEBUG
                    cout << "N or B vs Pawns (" << ColorImage(side) <<
                      ") = " << trade_score << endl;
#endif
                }
            }
            else if (ourQ != oppQ) {
              // penalize Q traded for Rook + Knight or bishop
              if (ourQ == 0) {
                 if (ourR==oppR+1 && ourMinors==oppMinors+1) {
                    // penalize Q traded for Rook + Knight or bishop
                    trade_score -= PAWN_VALUE/2;
                 }
                 else if (ourR==oppR && ourMinors==oppMinors+2) {
                    // Queen vs. 2 minors + pawns 
                    trade_score -= PAWN_VALUE/2;
                 }
#ifdef EVAL_DEBUG
                 cout << "QvsRB or QvsRN (" << ColorImage(side) <<
                   ") = " << trade_score << endl;
#endif
              }
            }
            else {                                // rooks un-equal
                // 2 minors vs. Rook. Slight bonus for minors (worth > 1
                // pawn over Rook).
                if (ourMinors == oppMinors+2) {   // rook vs. 2 minors
                    trade_score += RBN_TRADE[Util::Min(7,oppmat.pieceCount())];
                }
                // rook vs. minor
                else if (ourR+1==oppR && ourMinors==oppMinors+1) {
                   // minor vs. Rook, w. pawns
                   // side with minor actually gets bonus. Because, per
                   // Kaufman, this is less good for the Rook side
                   // than material balance alone would indicate. But we
                   // give a bit larger score for the Rook than Kaufman.
                   trade_score += RB_TRADE[Util::Min(7,ourmat.pieceCount())];
                }
#ifdef EVAL_DEBUG
                cout << "minors trade score (" << ColorImage(side) <<
                 ") = " << trade_score << endl;
#endif
            }
        }
        else {                                    // Q vs. 2 R
            if (oppmat.queenCount() == ourmat.queenCount()+1 &&
                ourmat.rookCount() == oppmat.rookCount()+2) {
                if (!(oppmat.pieceBits() == Material::KQ &&
                      ourmat.pieceBits() == Material::KRR)) {
                    trade_score -= PAWN_VALUE;    // bonus for queen if other pieces
                }
                else
                   trade_score -= PAWN_VALUE/3;  // less bonus for queen
#ifdef EVAL_DEBUG
                 cout << "queen trade score (" << ColorImage(side) <<
                 ") = " << trade_score << endl;
#endif
            }
        }
    }
    else {
        int mat_total = oppmat.materialLevel()+ourmat.materialLevel();
        // this catches cases where material value is even but a piece
        // has been traded for pawns (usually a bad idea unless pawns
        // are far advanced). It also gives a bonus for large material
        // imbalance.
        int bonus = (ourmat.materialLevel()-oppmat.materialLevel())*8;
        if (bonus<-50) bonus = -50;
        else if (bonus>50) bonus = 50;
        if (mat_total < 32) {
            if (bonus>0) bonus = (bonus+32-mat_total);
            else bonus = (bonus-32+mat_total);
        }
        score += bonus;
#ifdef EVAL_DEBUG
        cout << "score material adjust (" << ColorImage(side) <<
            ") = " << bonus << endl;
#endif
    }
    score += trade_score;
    return score;
}


static int adjustMaterialScoreNoPawns( const Board &board, ColorType side )
{
    // pawnless endgames. Generally drawish in many cases.
    const Material &ourmat = board.getMaterial(side);
    const Material &oppmat = board.getMaterial(OppositeColor(side));
    int score = 0;
    if (ourmat.infobits() == Material::KQ) {
        if (oppmat.infobits() == Material::KRR) {
            score -= (QUEEN_VALUE-2*ROOK_VALUE);  // even
        }
        else if (oppmat.infobits() == Material::KRB) {
                                                  // even
            score -= QUEEN_VALUE-(ROOK_VALUE+BISHOP_VALUE);
        }
                                                  // close to even
        else if (oppmat.infobits() == Material::KRN) {
            score -= QUEEN_VALUE-(ROOK_VALUE+BISHOP_VALUE)+PAWN_VALUE/4;
        }
    }
    else if (ourmat.infobits() == Material::KR) {
        if (oppmat.infobits() == Material::KBN) { // close to even
            score += (KNIGHT_VALUE+BISHOP_VALUE-ROOK_VALUE)-PAWN_VALUE/4;
        }
                                                  // even
        else if (oppmat.infobits() == Material::KB) {
            score -= (ROOK_VALUE-BISHOP_VALUE);
        }
                                                  // close to even
        else if (oppmat.infobits() == Material::KN) {
            score -= (ROOK_VALUE-KNIGHT_VALUE)+PAWN_VALUE/4;
        }
    }
    else if (ourmat.infobits() == Material::KRR) {
        if (oppmat.infobits() == Material::KRB) { // even
            score -= ROOK_VALUE-BISHOP_VALUE;
        }
                                                  // close to even
        else if (oppmat.infobits() == Material::KRN) {
            score -= ROOK_VALUE-KNIGHT_VALUE-PAWN_VALUE/4;
        }
                                                  // close to even
        else if (oppmat.infobits() == Material::KRBN) {
            score += (KNIGHT_VALUE+BISHOP_VALUE-ROOK_VALUE)-PAWN_VALUE/4;
        }
    }
    return score;
}


int Scoring::materialScore( const Board &board)
{
    // Computes material score, from the perspective of 'side'.

    ColorType side = board.sideToMove();
    const Material &ourmat = board.getMaterial(side);
    const Material &oppmat = board.getMaterial(OppositeColor(side));
    int mdiff =  (int)(ourmat.value() - oppmat.value());
    int score = 0;
#ifdef EVAL_DEBUG
    cout << "mdiff=" << mdiff << endl;
#endif
    const int our_pieces = ourmat.pieceBits();
    const int opp_pieces = oppmat.pieceBits();
    // check for bishop endgame - drawish
    if (our_pieces == Material::KB && opp_pieces == Material::KB) {
        if (Util::Abs(ourmat.pawnCount() - oppmat.pawnCount()) < 4)
            mdiff -= mdiff/4;
    }
    if (our_pieces != opp_pieces) {
#ifdef EVAL_DEBUG
        int tmp = score;
#endif
        if (ourmat.noPawns() && oppmat.noPawns()) {
            score += adjustMaterialScoreNoPawns(board,side) -
                adjustMaterialScoreNoPawns(board,OppositeColor(side));
        }
        else {
            score += adjustMaterialScore(board,side) -
                adjustMaterialScore(board,OppositeColor(side));
        }
#ifdef EVAL_DEBUG
        if (score-tmp) cout << "material score adjustment = " << score-tmp << endl;
#endif
    }
    return score+mdiff;
}


// Calculate a king cover score
int Scoring::calcCover(const Board &board, ColorType side, Square kp)
{
    int cover = -60;
    if (side == White) {
        Bitboard cover0(kingCoverW0[kp]);
        cover0 &= board.pawn_bits[White];
        cover += cover0.bitCount()*KING_COVER0;
        Bitboard cover1(kingCoverW1[kp]);
        cover1 &= board.pawn_bits[White];
        cover += cover1.bitCount()*KING_COVER1;
        Bitboard cover2(kingCoverW2[kp]);
        cover2 &= board.pawn_bits[White];
        cover += cover2.bitCount()*KING_COVER2;
    }
    else {
        Bitboard cover0(kingCoverB0[kp]);
        cover0 &= board.pawn_bits[Black];
        cover += cover0.bitCount()*KING_COVER0;
        Bitboard cover1(kingCoverB1[kp]);
        cover1 &= board.pawn_bits[Black];
        cover += cover1.bitCount()*KING_COVER1;
        Bitboard cover2(kingCoverB2[kp]);
        cover2 &= board.pawn_bits[Black];
        cover += cover2.bitCount()*KING_COVER2;
    }
    cover = Util::Min(0,cover);
    return cover;
}


void Scoring::calcCover(const Board &board,ColorType side,
                        KingCoverHashEntry &cover)
{
    Square kp = board.kingSquare(side);
    cover.cover = 0;
    int kpfile = File(kp);
    int k_cover = 0, q_cover = 0;
    int kprank = Rank(kp,side);
    int king_out = 0;
    if (kprank > 1)
        king_out = KING_OFF_BACK_RANK;
    if (kprank > 2)
        king_out *= 2;
    cover.cover += king_out;
    if (kpfile > 5) {
       k_cover = calcCover(board,side,kp);
#ifdef EVAL_DEBUG
        cout << "k-side king pawn shelter (" << ColorImage(side) << ") = " << k_cover << endl;
#endif
        for (int i=6;i<=8;i++) {
            if (FileOpen(board,i))
                k_cover += KING_NEAR_OPEN_FILE;
            else if (FileOpen(board,i,side))
                k_cover += KING_NEAR_HALF_OPEN_FILE;
        }
        cover.cover = k_cover;
    }
    else if (kpfile < 4) {
       q_cover = calcCover(board,side,kp);
#ifdef EVAL_DEBUG
        cout << "q-side king pawn shelter (" << ColorImage(side) << ") = " << k_cover << endl;
#endif
        for (int i=1;i<=3;i++) {
            if (FileOpen(board,i))
                q_cover += KING_NEAR_OPEN_FILE;
            else if (FileOpen(board,i,side))
                q_cover += KING_NEAR_HALF_OPEN_FILE;
        }
        cover.cover = q_cover;
    }
    else {
       Square ks = MakeSquare(GFILE,Rank(kp,side),side);
       k_cover = calcCover(board,side,ks);
       ks = MakeSquare(BFILE,Rank(kp,side),side);
       q_cover = calcCover(board,side,ks);
       cover.cover = (k_cover + q_cover + calcCover(board,side,kp))/3;
       for (int i=3;i<=6;i++) {
         if (FileOpen(board,i))
            cover.cover += KING_NEAR_OPEN_FILE;
         else if (FileOpen(board,i,side))
            cover.cover += KING_NEAR_HALF_OPEN_FILE;
       }
    }
#ifdef EVAL_DEBUG
    cout << "king cover (" << ColorImage(side) << ") = " << cover.cover << endl;
#endif

    // check for enemy pawns near the king
    cover.pawn_proximity = 0;
    int count = 0;
    if (side == White) {
        Bitboard patcks(w_pawn_attacks[kp]&board.pawn_bits[Black]);
        Square sq;
        while (patcks.iterate(sq)) {
           ASSERT(Rank(sq,White)-Rank(kp,White)>=0);
           int prank = Rank(sq,White);
           int rankdiff = prank-Rank(kp,White);
           int x = KING_PAWN_PROXIMITY[rankdiff];
           if (x && rankdiff<=3) {
               ++count;
               if (Attacks::pawn_attacks[sq][White] & board.pawn_bits[White]) {
                 ++x;
               }
               else if (board[sq-8]==WhitePawn) {
                 --x;
               }
           }
#ifdef EVAL_DEBUG
           cout << "pawn prox (to W. king), sq " << SquareImage(sq) << " =" << x << endl;
#endif
           cover.pawn_proximity += x;
        }
        if (count)
          cover.pawn_proximity += (count-1);
    }
    else {
        Bitboard patcks(b_pawn_attacks[kp]&board.pawn_bits[White]);
        Square sq;
        while (patcks.iterate(sq)) {
           ASSERT(Rank(sq,Black)-Rank(kp,Black)>=0);
           int prank = Rank(sq,Black);
           int rankdiff = prank-Rank(kp,Black);
           int x = KING_PAWN_PROXIMITY[rankdiff];
           if (x && rankdiff<=3) {
              ++count;
              if (Attacks::pawn_attacks[sq][Black] & board.pawn_bits[Black]) {
                ++x;
              }
              else if (board[sq+8]==BlackPawn) {
                --x;
              }
           }
#ifdef EVAL_DEBUG
           cout << "pawn prox (to B. king), sq " << SquareImage(sq) << " =" << x << endl;
#endif
          cover.pawn_proximity += x;
       }
       if (count) 
         cover.pawn_proximity += (count-1);
    }
}

static int isPinned(const Board &board,Square sq,Square pinned,Square okp) {
  if (Attacks::directions[pinned][okp] == Attacks::directions[sq][pinned]) {
    return board.clear(pinned,okp);
  }
  else
    return 0;
}


int Scoring::pieceScore(const Board &board,ColorType side,GamePhase phase,
const PawnHashEntry::PawnData &ourPawnData,
const PawnHashEntry::PawnData &oppPawnData,
int cover, int proximity)
{
    Bitboard b(board.occupied[side] & ~board.pawn_bits[side]);
    b.clear(board.kingSquare(side));
    int score = 0;
    int attack_count = 0;
    static const int ATTACK_FACTOR[6] = {0,1,2,2,3,4};
    static const int STACK = 2;
    int pin_count = 0;
    Square okp = board.kingSquare(OppositeColor(side));
    Bitboard nearKing(Attacks::king_attacks[okp]);
    nearKing.set(okp);
    Bitboard attackMask(nearKing & (board.allOccupied |
                                    oppPawnData.opponent_pawn_attacks));
    ColorType oside = OppositeColor(side);
    if (phase < ENDGAME_TRANSITION) {
       if (side == White) {
          if (board[D2] == WhitePawn && board[D3] > WhitePawn &&
                                         board[D3] < BlackPawn) {
             score += CENTER_PAWN_BLOCK;
          }
          if (board[E2] == WhitePawn && board[E3] > WhitePawn &&
                                       board[E3] < BlackPawn) {
             score += CENTER_PAWN_BLOCK;
          }
       }
       else {
          if (board[D7] == BlackPawn && board[D6] > BlackPawn) {
             score += CENTER_PAWN_BLOCK;
          }
          if (board[E7] == BlackPawn && board[E6] > BlackPawn) {
             score += CENTER_PAWN_BLOCK;
          }
       }
    }
    Square sq;
    while (b.iterate(sq)) {
        Square scoreSq = (side == White) ? sq : 63-sq;
#ifdef EVAL_DEBUG
        int tmp = score;
#endif
        switch(TypeOfPiece(board[sq])) {
            case Knight:
            {
                switch(phase) {
                    case OPENING:
                    case MIDDLEGAME:
                        score += KnightScores[scoreSq]; break;
                    case ENDGAME_TRANSITION:
                        score += KnightScores[scoreSq]/2; break;
                    case ENDGAME:
                        break;
                }
#ifdef EVAL_DEBUG
                cout << "Knight center score (" << SquareImage(sq) << "): ";
                cout << score-tmp << endl;
                int tmp1 = score;
#endif
                Bitboard knattacks(Attacks::knight_attacks[sq]);
                int mobl = Bitboard(knattacks &
                  ~board.occupied[side] & 
                ((~board.allOccupied & ~ourPawnData.opponent_pawn_attacks) |
                 (board.occupied[oside] & ~board.pawn_bits[oside]))
                 ).bitCount();
                ASSERT(mobl < 9);
                score += KNIGHT_MOBILITY[mobl];
#ifdef EVAL_DEBUG
                cout << "Knight mobility (" << SquareImage(sq) << "): ";
                cout << score-tmp1 << endl;
#endif
                int outpost = 0;
                if (side == White ) {
                  if (OutpostScoresW[sq] &&
                    !TEST_MASK(outpostW[sq],board.pawn_bits[Black])) {
                    outpost = OutpostScoresW[sq];
                  }
                }
                else {
                   if (OutpostScoresB[sq] &&
                     !TEST_MASK(outpostB[sq],board.pawn_bits[White])) {
                     outpost = OutpostScoresB[sq];
                   }
                }
                if (outpost) {
                   int outpostScore = outpost/2;
                   if (oppPawnData.opponent_pawn_attacks.isSet(sq)) {
                      // defended by pawns
                      outpostScore += outpost;
                   }
                   if (phase == ENDGAME) {
                     outpostScore /= 2;
                   }
                   score += outpostScore;
#ifdef EVAL_DEBUG
                   cout << "Knight outpost (" << SquareImage(sq) << "): ";
                   cout << outpostScore << endl;
#endif
                }

                if (phase != ENDGAME) {
                  Bitboard attacks(knattacks & nearKing);
                  if (!attacks.is_clear()) {
                    attack_count += ATTACK_FACTOR[Knight];
                    attackMask |= knattacks;
                  }
                }
#ifdef EVAL_DEBUG
                cout << ColorImage(side) << " Knight on " <<
                  SquareImage(sq) << " : " << score-tmp << endl;
#endif
                break;
            }
            case Bishop:
            {
#ifdef EVAL_DEBUG
                int tmp1 = score;
#endif
                switch(phase) {
                    case OPENING:
                    case MIDDLEGAME:
                        score += BishopScores[scoreSq]; break;
                    case ENDGAME_TRANSITION:
                        score += BishopScores[scoreSq]/2; break;
                    case ENDGAME:
                        break;
                }
                const Bitboard battacks(board.bishopAttacks(sq));
                if (phase != ENDGAME) {
                   Bitboard battacks2(board.bishopAttacks(sq,side));
                   Bitboard attacks(battacks2 & nearKing);
                   if (!attacks.is_clear()) {
                      attack_count += ATTACK_FACTOR[Bishop];
                      attackMask |= attacks;
                      if (Bitboard(battacks & nearKing) != Bitboard(battacks2 & nearKing)) {
                         // stacked attackers
#ifdef EVAL_DEBUG
                         cout << "stack=" << STACK << endl;
#endif 
                         attack_count += STACK;
                      }
                   }
                   Bitboard b(battacks & 
                              (board.knight_bits[oside] | board.rook_bits[oside]));
                   Square pinned;
                   while (b.iterate(pinned)) {
                      if (isPinned(board,sq,pinned,okp)) {
                          pin_count++;
                          break;
                       }
                   }
                }
#ifdef EVAL_DEBUG
                int tmp = score;
#endif
                score += BISHOP_MOBILITY*(Util::Min(6,Bitboard(battacks & ~board.allOccupied  & ~ourPawnData.opponent_pawn_attacks).bitCount()-4));
#ifdef EVAL_DEBUG
                cout << "bishop mobility (" << SquareImage(sq) <<
                  "): " << score-tmp << endl;
                cout << ColorImage(side) << " Bishop on " <<
                  SquareImage(sq) << " : " << score-tmp1 << endl;
#endif
            }
            break;
            case Rook:
            {
                int file = File(sq);
                const Bitboard rattacks(board.rookAttacks(sq));
                int rank = Rank(sq,side);
#ifdef EVAL_DEBUG
                int tmp = score;
#endif            


                if (rank == 7) {
                   int pawns_on_2nd_rank;
                   if (side == White) {
                      pawns_on_2nd_rank = TEST_MASK(Attacks::rank_mask[1],board.pawn_bits[White]);
                   }
                   else {
                      pawns_on_2nd_rank = TEST_MASK(Attacks::rank_mask[6],board.pawn_bits[Black]);
                   }
                   if (Rank(okp,oside)<=2 || pawns_on_2nd_rank) {
                      score += ROOK_ON_7TH_RANK;    
                      if (!Bitboard(rattacks & 
                          Attacks::rank_mask[Rank(sq,White)-1] &
                          board.rook_bits[side]).is_clear()) {
                          // 2 connected rooks on 7th
                          score += 2*ROOK_ON_7TH_RANK;
#ifdef EVAL_DEBUG
                          cout << "rook(s) on 7th (" << ColorImage(side) << "): " << score-tmp << endl;
#endif            
                      }
                   }
                }
                if (FileOpen(board,file,side) &&
                    (oppPawnData.weakFile(file) || FileOpen(board,file))) {
                   score += ROOK_ON_OPEN_FILE;
#ifdef EVAL_DEBUG
                      cout << "rook on weak/open file (" << ColorImage(side) << "): " << ROOK_ON_OPEN_FILE << endl;
#endif            
                   if (!Bitboard(rattacks & Attacks::file_mask_up[sq] &
                                 board.rook_bits[side]).is_clear()) {
                      // doubled rooks on weak or open file
                      score += DOUBLED_ROOKS;
#ifdef EVAL_DEBUG
                      cout << "doubled rooks (" << ColorImage(side) << "): " << DOUBLED_ROOKS << endl;
#endif            
                   }
                }
                Bitboard rattacks2(board.rookAttacks(sq,side));
                int mobl = Bitboard(rattacks2 & ~board.allOccupied
                  & ~ourPawnData.opponent_pawn_attacks).bitCount();
                // penalize low Rook mobility
                ASSERT(mobl < 16);
                score += ROOK_MOBILITY[mobl];
#ifdef EVAL_DEBUG
                cout << "rook mobility: " << mobl <<  " score=" << 
                  ROOK_MOBILITY[mobl] << endl;
#endif            
                if (phase != ENDGAME) {
                  Bitboard attacks(rattacks2 & nearKing);
                  if (!attacks.is_clear()) {
                    attack_count += ATTACK_FACTOR[Rook];
                    attackMask |= attacks;
                    if (Bitboard(rattacks & nearKing) != Bitboard(rattacks2 & nearKing)) {
                       // stacked attackers
#ifdef EVAL_DEBUG
                       cout << "stack=" << STACK << endl;
#endif
                       attack_count += STACK;
                    }
                  }
                  Bitboard b(rattacks & 
                             (board.bishop_bits[oside] | board.knight_bits[oside]));
                  Square pinned;
                  while (b.iterate(pinned)) {
                    if (isPinned(board,sq,pinned,okp)) {
                       pin_count++;
                       break;
                     }
                  }
                }
#ifdef EVAL_DEBUG
                cout << ColorImage(side) << " Rook on " <<
                  SquareImage(sq) << " : " << score-tmp << endl;
#endif
                break;
            }
            case Queen:
            {
                if (phase != ENDGAME) {
                   if (phase == OPENING && Rank(sq,side)>3) {
                       Bitboard back((board.knight_bits[side] | board.bishop_bits[side]) &
                        Attacks::rank_mask[side == White ? 0 : 7]);
#ifdef EVAL_DEBUG
                       int tmp = score;
#endif
                       score += QUEEN_OUT*(int)back.bitCount();
#ifdef EVAL_DEBUG
                       if (score != tmp) cout << "queen out: " << score-tmp << endl;
#endif
                    }
                    int any_attacks = 0;
                    Bitboard qattacks(board.bishopAttacks(sq));
                    Bitboard qattacks2(qattacks & nearKing);
                    if (!qattacks2.is_clear()) {
                       attack_count += ATTACK_FACTOR[Queen];
                       attackMask |= qattacks2;
                       any_attacks++;
                    }
                    Bitboard b(qattacks & 
                               (board.knight_bits[oside] | board.rook_bits[oside]));
                    Square pinned;
                    while (b.iterate(pinned)) {
                       if (board.isPinned(oside,sq,pinned,okp)) {
                           pin_count++;
                           break;
                        }
                    }
                    qattacks = Bitboard(board.rookAttacks(sq));
                    qattacks2 = Bitboard(qattacks & nearKing);
                    if (!qattacks2.is_clear()) {
                       attackMask |= qattacks2;
                       if (!any_attacks) {
                         attack_count += ATTACK_FACTOR[Queen];
                       }
                       else
                         attack_count++; // Queen attacks from 2 dirs
                    }
                    else if (qattacks & board.rook_bits[side]) {
                       if (board.rookAttacks(sq,side) & nearKing) {
                         attack_count += ATTACK_FACTOR[Queen];
                       }
                    }
                    b = (qattacks & (board.knight_bits[oside] | board.bishop_bits[oside]));
                    while (b.iterate(pinned)) {
                       if (board.isPinned(oside,sq,pinned,okp)) {
                           pin_count++;
                           break;
                       }
                    }
                }
#ifdef EVAL_DEBUG
                cout << ColorImage(side) << " Queen on " <<
                  SquareImage(sq) << " : " << score-tmp << endl;
#endif
                break;
            }
            default:
                break;
        }
    }
    if (board.getMaterial(side).bishopCount() == 1) {
      if (TEST_MASK(white_squares,board.bishop_bits[side])) {
        int whitePawns = ourPawnData.w_square_pawns;
        int blackPawns = ourPawnData.b_square_pawns;
        if ((whitePawns + blackPawns > 4) && (whitePawns > blackPawns+2)) {
#ifdef EVAL_DEBUG
            int tmp = score;
#endif
            score += BAD_BISHOP*(whitePawns-blackPawns-2);
#ifdef EVAL_DEBUG
            cout << "bad bishop (" << ColorImage(side) << ") : " << score-tmp << endl;
#endif
        }
      } 
      else {
        int whitePawns = ourPawnData.w_square_pawns;
        int blackPawns = ourPawnData.b_square_pawns;
        if ((whitePawns + blackPawns > 4) && (blackPawns > whitePawns+2)) {
#ifdef EVAL_DEBUG
            int tmp = score;
#endif
            score += BAD_BISHOP*(blackPawns-whitePawns-2);
#ifdef EVAL_DEBUG
            cout << "bad bishop (" << ColorImage(side) << ") : " << score-tmp << endl;
#endif
        }
      }
    }
    if (phase != ENDGAME) {
        int escape = 0;
#ifdef EVAL_DEBUG
        cout << "attack count=" << attack_count << endl;
#endif
        if (attack_count) {
           if (proximity>=3 && attack_count>=3) {
             proximity++;
             if (proximity >=6 || attack_count >= 6) {
               proximity++;
             }
           }
#ifdef EVAL_DEBUG
           cout << "attack count after proximity adjust=" << attack_count << endl;
#endif
           escape = Bitboard(Attacks::king_attacks[okp] & ~attackMask).bitCount();
           switch(escape) {
           case 0:
             attack_count+=2;
             break;
           case 1:
             attack_count++;
             break;
           case 2: break;
           default:
             attack_count--;
             break;
           }
#ifdef EVAL_DEBUG
           cout << "attack count after escape adjust=" << attack_count << endl;
#endif
           attack_count = Util::Min(attack_count,23);
        }
#ifdef EVAL_DEBUG
        cout << ColorImage(side) << " piece attacks on opposing king:" << endl;
        cout << " cover= " << cover << endl;
        cout << " attack_count=" << attack_count << endl;
        cout << " pawn proximity=" << proximity << endl;
        cout << " pin_count=" << pin_count << endl;
        cout << " escape squares=" << escape << endl;
#endif
        int attack = ATTACK_COUNT_SCALE[attack_count] + 
          PIN_MULTIPLIER*pin_count + PROXIMITY_MULTIPLIER*proximity;
        int boost = 0;
#ifdef BOOST
        int cover1 = cover/2;
        if (cover1 < -KING_ATTACK_BOOST_THRESHOLD && phase < ENDGAME_TRANSITION) {
           boost = Util::Min(KING_ATTACK_MAX_BOOST,-attack*(cover1+KING_ATTACK_BOOST_THRESHOLD)/96);
        }
#endif
        int kattack = attack - cover + boost;
#ifdef EVAL_DEBUG
        if (boost) cout << " king attack boost (" << ColorImage(side) <<
                     ") : " << boost << endl;
        cout << " pre-scaled king attack score (" << ColorImage(side) <<
          ") : " << kattack << endl;
        int level = board.getMaterial(side).materialLevel();
        cout << " level=" << level;
        cout << " scale factor=" << KING_ATTACK_MATERIAL_SCALE[level] << endl;
#endif

        kattack = (kattack*KING_ATTACK_MATERIAL_SCALE[board.getMaterial(side).materialLevel()])/32;
#ifdef EVAL_DEBUG
        cout << " scaled king attack score (" << ColorImage(side) << ") : " << kattack << endl;
#endif
      score += kattack;
    }
    return score;
}


int Scoring::calcPawnData(const Board &board, ColorType side,
PawnHashEntry::PawnData &entr)
{
    // This function computes a pawn structure score that depends
    // only on the location of pawns (of both colors). It also fills
    // in the pawn hash entry.
    //
    memset(&entr,'\0',sizeof(PawnHashEntry::PawnData));
    entr.first_passer = 9;
    entr.last_passer = 0;

    int score = 0;
    int incr = (side == White) ? 8 : -8;
    const ColorType oside = OppositeColor(side);
    entr.opponent_pawn_attacks = board.allPawnAttacks(oside);
    Piece enemy_pawn = MakePiece(Pawn,oside);
    Piece our_pawn = MakePiece(Pawn,side);
    Bitboard bi(board.pawn_bits[side]);
    Bitboard all_pawns(board.pawn_bits[White] | board.pawn_bits[Black]);
    Square sq;
    int lside = 0, rside = 0;
#ifdef PAWN_DEBUG
    int total_space = 0;
#endif
    while (bi.iterate(sq)) {
#ifdef PAWN_DEBUG
        int tmp = score + entr.passer_score;
        cout << ColorImage(side) << " pawn on " << FileImage(sq) <<
            RankImage(sq);
#endif
        int file = File(sq);
        int rank = Rank(sq,side);
        if (file < 4) {
            lside++;
        }
        else if (file > 5) {
            rside++;
        }
        if (SquareColor(sq) == White)
            entr.w_square_pawns++;
        else
            entr.b_square_pawns++;
        int backward = 0;
        int passed = 0;
        int weak = 0;
        int doubled = 0;
        int isolated;
        Bitboard doubles;
        if (side == White) {
            backward = !TEST_MASK(board.pawn_bits[side],backwardW[sq]);
            passed = !TEST_MASK(board.pawn_bits[oside],passedW[sq]);
            doubles = (board.pawn_bits[side] & Attacks::file_mask_up[sq]);
        }
        else {
            backward = !TEST_MASK(board.pawn_bits[side],backwardB[sq]);
            passed = !TEST_MASK(board.pawn_bits[oside],passedB[sq]);
            doubles = (board.pawn_bits[side] & Attacks::file_mask_down[sq]);
        }
#ifdef PAWN_DEBUG
        if (backward) cout << " backward";
#endif
        if (file == AFILE) {
            isolated = !OnFile(board.pawn_bits[side],BFILE);
        }
        else if (file == HFILE) {
            isolated = !OnFile(board.pawn_bits[side],GFILE);
        }
        else {
            isolated = !OnFile(board.pawn_bits[side],file-1) &&
                !OnFile(board.pawn_bits[side],file+1);
        }
        if (!doubles.is_clear()) {
#ifdef PAWN_DEBUG
            cout << " doubled";
#endif
            passed = 0;                           // our own pawn is ahead of us, don't count as passed
            doubled++;
            score += DOUBLED_PAWNS;
            if (doubles.bitCount() > 1) score += TRIPLED_PAWNS;
        }
        if (isolated && !passed && !doubled) {
            entr.isolated_count++;
#ifdef PAWN_DEBUG
            cout << " isolated";
#endif
        }
        if (rank<6) {
            int defenders = pawn_attacks(board,sq,side).bitCount();
            if (board[sq+incr] != enemy_pawn) {
                if (defenders == 0) {
                    Square sq2(sq + incr);
                    int patcks = pawn_attacks(board,sq2,oside).bitCount() -
                        pawn_attacks(board,sq2,side).bitCount();
                    if (patcks > 0) {
#ifdef PAWN_DEBUG
                        cout << " weak";
#endif
                        score += WEAK;
                        weak++;
                    }
                    if (patcks > 1) {             // "extra weak" pawn
                        score += WEAK2;
#ifdef PAWN_DEBUG
                        cout << " extra weak";
#endif
                    }
                }
            }
        }
        if (!passed && (weak || backward || isolated)) {
            entr.weak_files = (byte)(entr.weak_files | (byte)(1<<(file-1)));
            if (!OnFile(board.pawn_bits[oside],file)) {
                entr.weakopen++;
#ifdef PAWN_DEBUG
                cout << " weak/open";
#endif
            }
        }
        if (passed) {
#ifdef PAWN_DEBUG
            cout << " passed";
#endif
            ASSERT(rank<8);
            if (file > entr.last_passer)
                entr.last_passer = file;
            if (file < entr.first_passer)
                entr.first_passer = file;
            entr.passer_score += PASSED_PAWN[rank];
            if (side == White) {
              if (TEST_MASK(Attacks::pawn_attacks[sq][White],
                            board.pawn_bits[White]))
                 entr.flags &= PawnHashEntry::PROTECTED_PP;
            }
            else {
              if (TEST_MASK(Attacks::pawn_attacks[sq][Black],
                            board.pawn_bits[Black]))
                entr.flags &= PawnHashEntry::PROTECTED_PP;
            }
            entr.passers.set(sq);
        }
        else {
          int potential = 0;
          // not passed, check for potential passer
          Square blocker;
          if (side == White) {
             Bitboard ahead(board.pawn_bits[Black] & passedW[sq]);
             if (ahead.bitCount()==1) {
               // There is only one enemy pawn blocking us
               blocker = ahead.firstOne();
               int diff = (int)File(blocker)-(int)File(sq);
               if (diff != 0) {
                  potential = 1;
                  for (int file = File(sq)+diff;
                       file>=1 && file<=8 && potential;
                       file += diff) {
                     Bitboard bpawns(board.pawn_bits[Black] &
                                     Attacks::file_mask[file-1]);
                     if (bpawns.is_clear()) break; // file is clear
                     Square bpawn;
                     while (bpawns.iterate(bpawn)) {
                       // check each black pawn has a white blocker
                       if (!(Attacks::file_mask_down[bpawn] &
                             board.pawn_bits[White])) {
                          potential = 0;
                          break;
                       }
                     }
                  }
               }
             }
          }
          else {
             Bitboard ahead(board.pawn_bits[White] & passedB[sq]);
             if (ahead.bitCount()==1) {
               // There is only one enemy pawn blocking us
               blocker = ahead.firstOne();
               int diff = (int)File(blocker)-(int)File(sq);
               if (diff != 0) {
                  potential = 1;
                  for (int file = File(sq)+diff;
                       file>=1 && file<=8 && potential;
                       file += diff) {
                     Bitboard wpawns(board.pawn_bits[White] &
                                     Attacks::file_mask[file-1]);
                     if (wpawns.is_clear()) break; // file is clear
                     Square wpawn;
                     while (wpawns.iterate(wpawn)) {
                       // check each White pawn has a Black blocker
                       if (!(Attacks::file_mask_up[wpawn] &
                             board.pawn_bits[Black])) {
                          potential = 0;
                          break;
                       }
                     }
                  }
               }
             }
          }
        }
        int duo = 0;
        if (rank > 3 && file != 8 && board[sq+1] == our_pawn) {
#ifdef PAWN_DEBUG
            cout << " duo";
#endif
            duo++;
        }
        if (!passed && rank >= 4) {
            int space = SPACE*(rank-3);
            if (duo) space *=2;
            score += space;
#ifdef PAWN_DEBUG
            cout << " space=" <<space;
#endif
        }
#ifdef PAWN_DEBUG
        cout << " total = " << score+entr.passer_score-tmp << endl;
#endif
    }
    Bitboard passers2(entr.passers);
    while (passers2.iterate(sq)) {
       if (TEST_MASK(connected_passers[sq][side],entr.passers)) {
          entr.passer_score += CONNECTED_PASSERS[Rank(sq,side)];
#ifdef PAWN_DEBUG
          cout << "connected passer score, " 
               << SquareImage(sq) << " (";
          cout << ColorImage(side);
          cout << ") : " << CONNECTED_PASSERS[Rank(sq,side)];
          cout << endl;
#endif
       }
       else if (TEST_MASK(adjacent_passers[sq],entr.passers)) {
          entr.passer_score += ADJACENT_PASSERS;
#ifdef PAWN_DEBUG
          cout << "adjacent passer score (";
          cout << ColorImage(side);
          cout << ") : " << ADJACENT_PASSERS;
          cout << endl;
#endif
       }
    }
    Bitboard centerCalc(center & board.pawn_bits[side]);
    entr.center_score = PAWN_CENTER_SCORE*centerCalc.bitCount();
    centerCalc = Bitboard(center & board.allPawnAttacks(side));
    entr.center_score += PAWN_CENTER_SCORE2*centerCalc.bitCount();

    entr.score = score;
    return score;
}


void Scoring::evalOutsidePassers(const Board &board,
PawnHashEntry *pawnEntry)
{
    const int &first_passer_w = pawnEntry->wPawnData.first_passer;
    const int &last_passer_w = pawnEntry->wPawnData.last_passer;
    const int &first_passer_b = pawnEntry->bPawnData.first_passer;
    const int &last_passer_b = pawnEntry->bPawnData.last_passer;

    if (first_passer_w == 9 && first_passer_b == 9) {
        return;
    }
    Bitboard all_pawns(board.pawn_bits[White] | board.pawn_bits[Black]);
    int edge_dist_w_first = 9;
    int edge_dist_b_first = 9;
    int edge_dist_w_last = 9;
    int edge_dist_b_last = 9;
    if (first_passer_w <=4 && !TEST_MASK(all_pawns,left_side_mask[first_passer_w-1]) &&
        TEST_MASK(board.pawn_bits[White],efgh_mask) && TEST_MASK(board.pawn_bits[Black],efgh_mask)) {
        edge_dist_w_first = first_passer_w;
    }
    if (first_passer_b <= 4 && !TEST_MASK(all_pawns,left_side_mask[first_passer_b-1]) &&
       TEST_MASK(board.pawn_bits[Black],efgh_mask) && TEST_MASK(board.pawn_bits[White],efgh_mask)) {
       edge_dist_b_first = first_passer_b;
    }
    if (last_passer_w != 9 && last_passer_w >= 5 && !TEST_MASK(all_pawns,right_side_mask[last_passer_w-1]) &&
        TEST_MASK(board.pawn_bits[White],abcd_mask) && TEST_MASK(board.pawn_bits[Black],abcd_mask)) {
        edge_dist_w_last = 9-last_passer_w;
    }
    if (last_passer_b != 9 && last_passer_b >= 5 && !TEST_MASK(all_pawns,right_side_mask[last_passer_b-1]) &&
        TEST_MASK(board.pawn_bits[Black],abcd_mask) && TEST_MASK(board.pawn_bits[White],abcd_mask)) {
        edge_dist_b_last = 9-last_passer_b;
    }
    if (edge_dist_w_first<Util::Min(edge_dist_b_first,edge_dist_b_last)) {
       pawnEntry->wPawnData.flags &= PawnHashEntry::OUTSIDE_PP;
    }
    if (edge_dist_w_last<Util::Min(edge_dist_b_first,edge_dist_b_last)) {
       pawnEntry->wPawnData.flags &= PawnHashEntry::OUTSIDE_PP;
    }
    if (edge_dist_b_first<Util::Min(edge_dist_w_first,edge_dist_w_last)) {
       pawnEntry->bPawnData.flags &= PawnHashEntry::OUTSIDE_PP;
    }
    if (edge_dist_b_last<Util::Min(edge_dist_w_first,edge_dist_w_last)) {
       pawnEntry->bPawnData.flags &= PawnHashEntry::OUTSIDE_PP;
    }
    return;
}


void Scoring::calcPawnEntry(const Board &board, PawnHashEntry *pawnEntry)
{
    // pawn position not found, calculate the data we need
    calcPawnData(board,White,pawnEntry->wPawnData);
    calcPawnData(board,Black,pawnEntry->bPawnData);
    evalOutsidePassers(board,pawnEntry);
    pawnEntry->hc = board.pawnHash();
}


int Scoring::evalu8(const Board &board)
{
    return materialScore(board) + positionalScore(board,
        -Constants::MATE,Constants::MATE);
}


int Scoring::evalu8(const Board &board,int alpha,int beta)
{
    int mat = materialScore(board);
    return mat + positionalScore(board,alpha-mat,beta-mat);
}


int Scoring::pawnScore(const Board &board, ColorType side,
const PawnHashEntry::PawnData &pawnData, GamePhase phase)
{
    int score = pawnData.score;
#ifdef EVAL_DEBUG
    int tmp = score;
#endif
    switch(phase) {
    case OPENING:
        score += pawnData.center_score;
        break;
    case MIDDLEGAME:
        score += pawnData.center_score/2;
        break;
    default:
      break;
    }
#ifdef EVAL_DEBUG
    cout << "pawn center score (" << ColorImage(side) << "): " << score-tmp << endl;
    tmp = score;
#endif
    if (!(board.rook_bits[OppositeColor(side)] |
          board.queen_bits[OppositeColor(side)]).is_clear()) {
        score += (int)pawnData.weakopen*WEAK_ON_OPEN_FILE;

    }
#ifdef EVAL_DEBUG
    if( score-tmp)
      cout << ColorImage(side) << " weak/backward pawn penalty: " <<
        score - tmp << endl;
#endif
    int passer_score = pawnData.passer_score;
    // interaction of passed pawns and pieces
    ColorType oside = OppositeColor(side);
    Square sq;
    Bitboard passers2(pawnData.passers);
    while (passers2.iterate(sq)) {
       ASSERT(OnBoard(sq));
       int rank = Rank(sq,side);
       ASSERT(rank<8);
       Square sq2;
       if (side==White) sq2 = sq+8; else sq2 = sq-8;
       if (!IsEmptyPiece(board[sq2]) && PieceColor(board[sq2]) == oside) {
#ifdef PAWN_DEBUG
           int tmp = passer_score;
#endif
           passer_score -= PASSED_PAWN_BLOCK[rank];
#ifdef PAWN_DEBUG
           cout << ColorImage(side) << " pawn on " << SquareImage(sq) << " blocked, score=" << passer_score-tmp << endl;
#endif
       }
       int file = File(sq);
       if (TEST_MASK(board.rook_bits[side],Attacks::file_mask[file-1])) {
           Bitboard atcks(board.fileAttacks(sq) & board.rook_bits[side]);
           if (!atcks.is_clear()) {
              if (atcks & Attacks::file_mask_up[sq]) {
                 if (side == White) passer_score += ROOK_AHEAD_OF_PP;
                 else passer_score += ROOK_BEHIND_PP;
              }
              if (atcks & Attacks::file_mask_down[sq]) {
                 if (side == White) passer_score += ROOK_BEHIND_PP;
                 else passer_score += ROOK_AHEAD_OF_PP;
              }
          }
          // Rook adjacent to pawn on 7th is good too
          if (rank == 7 && file<8 && TEST_MASK(board.rook_bits[side],Attacks::file_mask[file])) {
             Bitboard atcks(board.fileAttacks(sq+1) & board.rook_bits[side]);
             if (!atcks.is_clear() || board.rook_bits[side].isSet(sq+1)) score += ROOK_BEHIND_PP;
          }
          if (rank == 7 && file>1 && TEST_MASK(board.rook_bits[side],Attacks::file_mask[file-2])) {
             Bitboard atcks(board.fileAttacks(sq-1) & board.rook_bits[side]);
             if (!atcks.is_clear() || board.rook_bits[side].isSet(sq-1)) score += ROOK_BEHIND_PP;
          }
       }
    }
#ifdef EVAL_DEBUG
    cout << ColorImage(side) << " passed pawn score: " << passer_score << endl;
#endif
    return score+passer_score;
}

Scoring::GamePhase Scoring::gamePhase(int level) {
    GamePhase phase;
    if (level < 13)
        phase = ENDGAME;
    else if (level < 16)
        phase = ENDGAME_TRANSITION;
    else if (level < 24)
        phase = MIDDLEGAME;
    else
        phase = OPENING;
    return phase;
}

int Scoring::scoreEndgame(const Board &board,
                          EndgameHashEntry *endgameEntry,
                          const PawnHashEntry::PawnData &pawnData,
                          ColorType side,
                          GamePhase phase) {
  int score = 0;
  /** TBD
  if (board.sideToMove() != side) {
      int x = opposition[board.kingSquare(board.sideToMove())]
        [board.kingSquare(board.oppositeSide())];
      if (x > 0) {
           score += OPPOSITION*x;
#ifdef EVAL_DEBUG
           cout << ColorImage(side) << " has opposition, score=" << OPPOSITION*x << endl;
#endif
      }
  }
  **/
  const ColorType oside = OppositeColor(side);
  if ((board.getMaterial(side).infobits() == Material::KBN) &&
    board.getMaterial(oside).kingOnly()) {
        // KBNK endgame, special case.
        Square oppkp = board.kingSquare(oside);
        Bitboard bishops(board.bishop_bits[side]);
        Square sq = bishops.firstOne();
        // try to force king to a corner where mate is possible.
        if (SquareColor(sq) == Black) {
            score += KBNKScores[Flip[oppkp]];
        }
        else {
            score += KBNKScores[oppkp];
        }
        // keep the kings close
        score += 10-distance1(board.kingSquare(White),board.kingSquare(Black));
#ifdef EVAL_DEBUG
        cout << "KBNK score (" << ColorImage(side) << ") = " << score << endl;
#endif
        return score;
  }
  else if ((board.getMaterial(side).pieceBits() == Material::KB &&
          board.getMaterial(oside).pieceBits() == Material::KN) ||
         (board.getMaterial(side).pieceBits() == Material::KBN &&
          board.getMaterial(oside).pieceBits() == Material::KNN)) {
    if (TEST_MASK(board.pawn_bits[oside],abcd_mask) &&
       TEST_MASK(board.pawn_bits[oside],efgh_mask)) {
       score += BISHOP_ENDGAME_BONUS;
    }
  }
  else if (board.getMaterial(side).pieceBits() == Material::KBB &&
              board.getMaterial(oside).pieceBits() == Material::KBN) {
    // both sides have bishops but one has BN vs BB
    if (TEST_MASK(board.pawn_bits[oside],abcd_mask) &&
       TEST_MASK(board.pawn_bits[oside],efgh_mask)) {
       score += BISHOP_ENDGAME_BONUS/2;
    }
  }
  if (side == White) {
    score = endgameEntry->wScore +
      (int)endgameEntry->white_endgame_pawn_proximity +
      (int)endgameEntry->white_king_position;
#ifdef EVAL_DEBUG
    cout << "king position (White) : " << (int)endgameEntry->white_king_position << endl;
    cout << "king/pawn proximity (White) : " << (int)endgameEntry->white_endgame_pawn_proximity << endl;
#endif
  }
  else {
    score = endgameEntry->bScore +
      (int)endgameEntry->black_endgame_pawn_proximity +
      (int)endgameEntry->black_king_position;
#ifdef EVAL_DEBUG
    cout << "king position (Black) : " << (int)endgameEntry->black_king_position << endl;
    cout << "king/pawn proximity (Black) : " << (int)endgameEntry->black_endgame_pawn_proximity << endl;
#endif
  }
  // King/Pawn interactions
  int uncatchables = (side == White) ? (int)endgameEntry->w_uncatchable :
    (int)endgameEntry->b_uncatchable;
  Bitboard passers2(pawnData.passers);
  Square passer;
  while (passers2.iterate(passer)) {
      // Encourage escorting a passer with the King, ideally with King in
      // front
      if (Util::Abs(File(board.kingSquare(side))-File(passer))<=1 && Rank(board.kingSquare(side),side)>=Rank(passer,side)) score += KING_AHEAD_OF_PAWN*Rank(passer,side);
  }
  if (uncatchables) {
     int pcount = board.getMaterial(OppositeColor(side)).pieceCount();
     int bonus = 0;
     if ((board.getMaterial(side).infobits() == Material::KP) &&
        board.getMaterial(oside).kingOnly()) {
        score += BITBASE_WIN;
     }
     else if (pcount == 0)
        score += uncatchables*UNCATCHABLE_PAWN*2;
     else if (pcount <=2)
        score += uncatchables*UNCATCHABLE_PAWN;
#ifdef EVAL_DEBUG
     cout << ColorImage(side) << " uncatchables=" << uncatchables << " bonus=" << bonus <<endl;
#endif
     score += bonus;
  }
  Bitboard rooks(board.rook_bits[side]);
  Square rooksq;
#ifdef EVAL_DEBUG
  int tmp = score;
#endif
  while (rooks.iterate(rooksq)) {
     if (TEST_MASK(board.pawn_bits[side],Attacks::rank_mask[Rank(rooksq,White)-1])) {
        Bitboard atcks(board.rankAttacks(rooksq) & board.pawn_bits[side]);
        Square pawnsq;
             int i = 0;
             while (atcks.iterate(pawnsq)) {
                if (!TEST_MASK(Attacks::pawn_attacks[pawnsq][side],
                    board.pawn_bits[side])) // not protected by pawn
                   score += SIDE_PROTECTED_PAWN*(i+1);
                ++i;
             }
     }
  }
#ifdef EVAL_DEBUG
  if (score-tmp)
        cout << ColorImage(side) << " rook side-protecting pawn: " << score-tmp << endl;
#endif

  switch(phase) {
  case OPENING:
  case MIDDLEGAME:
    return 0;
  case ENDGAME_TRANSITION:
    return score/2;
  case ENDGAME:
    return score;
  }
  return 0; // shouldn't get here
}


int Scoring::positionalScore( const Board &board, int alpha, int beta)
{
    EvalCacheEntry &cacheEntry = evalCache[board.hashCode() % EVAL_CACHE_SIZE];
    // copy from the cache
    hash_t key = cacheEntry.score_key;
    int score = cacheEntry.score;
    // xor key with values (trick from crafty) so when we retrieve we
    // will not fetch a value that does not have the score/best move
    // initially stored (which can happen as multiple threads write the cache).
    key ^= (hash_t)score;
    if (key == board.hashCode() && score != INVALID_SCORE) {
      if (score < Constants::MATE) {
         return score;
      }
      else {
        // a partial (lazy evaluated) score was in the cache
        score -= PARTIAL_SCORE_BIAS;
        if (score > beta + LAZY_MARGIN) {
           return score;
        }
        else if (score < alpha - LAZY_MARGIN) {
           return score;
        }
      }
    }

    // Position was not in cache, must evaluate it
    score = 0;
    int w_materialLevel = board.getMaterial(White).materialLevel();
    int b_materialLevel = board.getMaterial(Black).materialLevel();
    GamePhase wPhase = gamePhase(w_materialLevel);
    GamePhase bPhase = gamePhase(b_materialLevel);
#ifdef EVAL_DEBUG
    static const char *phases[4] = {"Opening","Middlegame","Endgame Transition","Endgame"};
    cout << "White material level=" << w_materialLevel << 
      " phase=" <<  phases[wPhase] << endl;
    cout << "Black material level=" << b_materialLevel << 
      " phase=" <<  phases[bPhase] << endl;
#endif

    hash_t pawnHash = board.pawnHashCodeW ^ board.pawnHashCodeB;

    PawnHashEntry *pawnEntry =
        &pawnHashTable[pawnHash % PAWN_HASH_SIZE];
    if (pawnEntry->hc != pawnHash) {
        calcPawnEntry(board,pawnEntry);
    }

#ifdef EVAL_DEBUG
    int tmp1 = score;
#endif
    score += pawnScore(board,White,pawnEntry->wPawnData,wPhase)-
        pawnScore(board,Black,pawnEntry->bPawnData,bPhase);
#ifdef EVAL_DEBUG
    int tmp = score;
#endif
    // outside passed pawn scoring, based on cached pawn data
    if (pawnEntry->wPawnData.outside() && !pawnEntry->bPawnData.outside()) {
        score += OUTSIDE_PP[board.getMaterial(Black).materialLevel()/2];
#ifdef EVAL_DEBUG
        cout << "outside pawn score (White) : " << score-tmp << endl;
#endif
    }
    else if (pawnEntry->bPawnData.outside() && !pawnEntry->wPawnData.outside()) {
        score -= OUTSIDE_PP[board.getMaterial(White).materialLevel()/2];
#ifdef EVAL_DEBUG
        cout << "outside pawn score (Black): " << score-tmp << endl;
#endif
    }
    // extra penalty for multiple isolated pawns
#ifdef PAWN_DEBUG
    tmp = score;
#endif
    if (pawnEntry->wPawnData.isolated_count) {
       if (board.getMaterial(Black).noPieces())
         score += ISOLATED_PAWN[pawnEntry->wPawnData.isolated_count]/2;
       else
         score += ISOLATED_PAWN[pawnEntry->wPawnData.isolated_count];
    }
#ifdef PAWN_DEBUG
    cout << "White isolated pawn penalty  = " << score-tmp << endl;
    tmp = score;
#endif
    if (pawnEntry->bPawnData.isolated_count) {
       if (board.getMaterial(White).noPieces())
          score -= ISOLATED_PAWN[pawnEntry->bPawnData.isolated_count]/2;
       else
          score -= ISOLATED_PAWN[pawnEntry->bPawnData.isolated_count];
    }
#ifdef PAWN_DEBUG
    cout << "Black isolated pawn penalty  = " << tmp-score << endl;
    tmp = score;
#endif

#ifdef EVAL_DEBUG
    cout << "net pawn score: " << score-tmp1 << endl;
    tmp = score;
#endif

    // Penalize loss of castling. Based on opponent material and whether
    // opponent has Queen
    if (wPhase != ENDGAME) {
       int castling = CASTLING_SCORES[(int)board.castleStatus(White)];
       int queenon = (castling < 0 && board.getMaterial(Black).queenCount()) ? 32 : 0;
       castling = ((castling-queenon)*board.getMaterial(Black).materialLevel())/32;                
#ifdef EVAL_DEBUG
       cout << "White castling score: " << castling << endl;
#endif
       score += castling;
    }
    if (bPhase != ENDGAME) {
       int castling = CASTLING_SCORES[(int)board.castleStatus(Black)];
       int queenon = (castling < 0 && board.getMaterial(White).queenCount()) ? 32 : 0;
       castling = ((castling-queenon)*board.getMaterial(White).materialLevel())/32;                
#ifdef EVAL_DEBUG
       cout << "Black castling score: " << -castling << endl;
#endif
       score -= castling;
    }
#ifdef EVAL_DEBUG
    cout << "net castling score: " << score-tmp << endl;
#endif

    int w_cover=0, w_proximity = 0, b_cover=0, b_proximity=0;
    if (Util::Min(wPhase,bPhase) < ENDGAME) {
        hash_t kcHash = BoardHash::kingCoverHash(board,White);
        KingCoverHashEntry *kingCoverEntryW =
            &kingCoverHashTable[kcHash % KING_COVER_HASH_SIZE];
        if (kingCoverEntryW->hc != kcHash) {
            calcCover(board,White,*kingCoverEntryW);
            kingCoverEntryW->hc = kcHash;
        }
        w_cover = kingCoverEntryW->cover;
        w_proximity = kingCoverEntryW->pawn_proximity;
        kcHash = BoardHash::kingCoverHash(board,Black);
        KingCoverHashEntry *kingCoverEntryB =
            &kingCoverHashTable[kcHash % KING_COVER_HASH_SIZE];
        if (kingCoverEntryB->hc != kcHash) {
            calcCover(board,Black,*kingCoverEntryB);
            kingCoverEntryB->hc = kcHash;
        }
        b_cover = kingCoverEntryB->cover;
        b_proximity = kingCoverEntryB->pawn_proximity;
    }
    if (board[A7] == WhiteBishop) {
        if (board[B6] == BlackPawn &&
        board.anyAttacks(B6,Black)) {
            score += BISHOP_TRAPPED;
        }
    }
    if (board[H7] == WhiteBishop) {
        if (board[G6] == BlackPawn &&
        board.anyAttacks(G6,Black)) {
            score += BISHOP_TRAPPED;
        }
    }
    if (board[A2] == BlackBishop) {
        if (board[B3] == WhitePawn &&
        board.anyAttacks(B3,White)) {
            score -= BISHOP_TRAPPED;
        }
    }
    if (board[H2] == BlackBishop) {
        if (board[G3] == WhitePawn &&
        board.anyAttacks(G3,White)) {
            score -= BISHOP_TRAPPED;
        }
    }
    int score1 = (board.sideToMove() == White) ? score : -score;
#ifdef LAZY_DEBUG
    int lazyHi = 0, lazyLo = 0;
    if (score1 > beta + LAZY_MARGIN) {
      lazyHi++;
    }
    else if (score1 < alpha - LAZY_MARGIN) {
      lazyLo++;
    }
#else
    if (score1 > beta + LAZY_MARGIN) {
       score = score1 + PARTIAL_SCORE_BIAS;
       cacheEntry.score_key = board.hashCode() ^ (hash_t)score;
       cacheEntry.score = score;
       return score1;
    }
    else if (score1 < alpha - LAZY_MARGIN) {
       score = score1 + PARTIAL_SCORE_BIAS;
       cacheEntry.score_key = board.hashCode() ^ (hash_t)score;
       cacheEntry.score = score;
       return score1;
    }
#endif
    score += pieceScore(board,White,wPhase,
        pawnEntry->wPawnData,
        pawnEntry->bPawnData,b_cover,b_proximity)-
        pieceScore(board,Black,bPhase,
        pawnEntry->bPawnData,
        pawnEntry->wPawnData,w_cover,w_proximity);

    if (Util::Max(wPhase,bPhase) >= ENDGAME_TRANSITION) {
       hash_t hc = BoardHash::kingPawnHash(board);
       EndgameHashEntry *endgameEntry = &endgameHashTable[hc % ENDGAME_HASH_SIZE];
       if (endgameEntry->hc != hc) {
         calcEndgame(board,pawnEntry,endgameEntry);
         endgameEntry->hc = hc;
       }
#ifdef EVAL_DEBUG
       tmp = score;
#endif
       score += scoreEndgame(board,endgameEntry,pawnEntry->wPawnData,White,wPhase) -
         scoreEndgame(board,endgameEntry,pawnEntry->bPawnData,Black,bPhase);
#ifdef EVAL_DEBUG
       cout << "net endgame score = " << score-tmp << endl;
#endif
    }
    if (board.sideToMove() == Black)
      score = -score;

    // Because positional scoring is inexact anyway, round the scores
    // so we will not change the selected move over a trivial difference.
    score = (score/4)*4;

#ifdef LAZY_DEBUG
    if (lazyLo) {
      if (score > alpha)
        cout << board << endl;
    }
    else if (lazyHi) {
      if (score < beta) 
        cout << board << endl;
    }
#endif
    // xor key with values (trick from crafty)
    cacheEntry.score_key = board.hashCode() ^ (hash_t)score;
    cacheEntry.score = score;

    return score;
}


static int KBPDrawW(const Board &board)
{
    Bitboard pb(board.pawn_bits[White]);
    if (OnFile(pb,AFILE)) {
        // on the a file
        // queen square is white
        if (!TEST_MASK(board.bishop_bits[White],white_squares)) {
            // wrong color bishop
            Square kp = board.kingSquare(Black);
            if (kp == A8 || Attacks::king_attacks[A8].isSet(kp)) {
                return 1 /*TRUE*/;
            }
        }
    }
    else if (OnFile(pb,HFILE)) {
        // on the h file
        // queen square is black
        if (!TEST_MASK(board.bishop_bits[White],black_squares)) {
            // wrong color bishop
            Square kp = board.kingSquare(Black);
            if (kp == H8 || Attacks::king_attacks[H8].isSet(kp)) {
                return 1 /*TRUE*/;
            }
        }
    }
    return 0;
}


static int KBPDrawB(const Board &board)
{
    Bitboard pb(board.pawn_bits[Black]);
    if (OnFile(pb,AFILE)) {
        // on the a file
        // queen square is black
        if (!TEST_MASK(board.bishop_bits[Black],black_squares)) {
            // wrong color bishop
            Square kp = board.kingSquare(White);
            if (kp == A1 || Attacks::king_attacks[A1].isSet(kp)) {
                return 1;
            }
        }
    }
    else if (OnFile(pb,HFILE)) {
        // on the h file
        // queen square is white
        if (!TEST_MASK(board.bishop_bits[Black],white_squares)) {
            // wrong color bishop
            Square kp = board.kingSquare(White);
            if (kp == H1 || Attacks::king_attacks[H1].isSet(kp)) {
                return 1;
            }
        }
    }
    return 0;
}


int Scoring::theoreticalDraw(const Board &board)
{
    const Material &mat1 = board.getMaterial(White);
    const Material &mat2 = board.getMaterial(Black);
    if (mat1.value() > KING_VALUE+(KNIGHT_VALUE*2) ||
        mat2.value() > KING_VALUE+(KNIGHT_VALUE*2))
        return 0;
    // Check for K + P vs rook pawn
    if (mat1.infobits() == Material::KP && mat2.kingOnly()) {
        if (TEST_MASK(rook_pawn_mask, board.pawn_bits[White])) {
            Square kp = board.kingSquare(White);
            Square kp2 = board.kingSquare(Black);
            Square psq = (board.pawn_bits[White].firstOne());
            return lookupBitbase(kp,psq,kp2,White,board.sideToMove())==0;
        }
    }
    else if (mat2.infobits() == Material::KP && mat1.kingOnly()) {
        if (TEST_MASK(rook_pawn_mask,board.pawn_bits[Black])) {
            Square kp = board.kingSquare(Black);
            Square kp2 = board.kingSquare(White);
            Square psq = (board.pawn_bits[Black].firstOne());
            return lookupBitbase(kp,psq,kp2,Black,board.sideToMove())==0;
        }
    }
    // Check for wrong bishop + rook pawn vs king.  Not very common but
    // we don't want to get it wrong.
    else if (mat1.infobits() == Material::KBP && mat2.kingOnly()) {
        return KBPDrawW(board);
    }
    else if (mat2.infobits() == Material::KBP && mat1.kingOnly()) {
        return KBPDrawB(board);
    }
    // check for KNN vs K
    if (((unsigned)mat1.infobits() == Material::KNN &&
        mat2.kingOnly()) ||
        ((unsigned)mat2.infobits() == Material::KNN &&
        mat1.kingOnly()))
        return 1;

    return 0;
}


int Scoring::repetitionDraw( const Board &board )
{
    return board.repCount() >= 2;
}


int Scoring::materialDraw( const Board &board)
{
    // check for insufficient material per FIDE Rules of Chess
    const Material &mat1 = board.getMaterial(White);
    const Material &mat2 = board.getMaterial(Black);
    if (mat1.pawnCount() || mat2.pawnCount()) {
        return 0 ;
    }
    if ((mat1.value() <= KING_VALUE + BISHOP_VALUE) &&
    (mat2.value() <= KING_VALUE + BISHOP_VALUE)) {
        if (mat1.kingOnly() || mat2.kingOnly())
            // K vs K, or K vs KN, or K vs KB
            return 1;
        else if (mat1.infobits() == Material::KN &&
        mat2.infobits() == Material::KN) {
            // KN vs KN
            return 0;
        }
        else {                                    /* KBKB */
            // drawn if same-color bishops
            if (TEST_MASK(board.bishop_bits[White],black_squares))
                return TEST_MASK(board.bishop_bits[Black],black_squares);
            else if (TEST_MASK(board.bishop_bits[White],white_squares))
                return TEST_MASK(board.bishop_bits[Black],white_squares);
        }
    }
    return 0 /*FALSE*/;
}


int Scoring::isLegalDraw(const Board &board)
{
    return (repetitionDraw(board) || materialDraw(board));
}


int Scoring::isDraw(const Board &board)
{
    return (repetitionDraw(board) || materialDraw(board) ||
        theoreticalDraw(board));
}


int Scoring::isDraw( const Board &board, int &rep_count, int ply)
{
    // First check for draw by repetition
    int entries = board.state.moveCount-2;
    rep_count = 0;
    // follow rule Crafty uses: 2 repeats if ply<=2, 1 otherwise:
    if (entries >= 0) {
        const int target = (ply<=2) ? 2 : 1;
        for (hash_t *rep_list=board.repListHead-3;
             entries>=0;
             rep_list-=2,entries-=2) {
            if (*rep_list == board.hashCode()) {
                rep_count++;
                if (rep_count >= target) {
                    return 1;
                }
            }
        }
    }
    // check the 50 move rule
    if (board.state.moveCount >= 100) {
        return 1;
    }
    if (board.getMaterial(White).value() <= KING_VALUE+(KNIGHT_VALUE*2) &&
    board.getMaterial(Black).value() <= KING_VALUE+(KNIGHT_VALUE*2)) {
        // check for insufficient material and other drawing situations
        return Scoring::isDraw(board);
    }
    else
        return 0;
}


void Scoring::printScore( int score, ostream &str)
{
    if (score <= -Constants::MATE+256) {
        str << "-Mate" << (Constants::MATE+score+1)/2;
    }
    else if (score >= Constants::MATE-256) {
        str << "+Mate" << (Constants::MATE-score+1)/2;
    }
    else {
      if (score >=0) str << '+';
      str << fixed << setprecision(2) << (score*1.0)/100.0;
    }
}


void Scoring::printScoreUCI( int score, ostream &str)
{
    if (score <= -Constants::MATE+256) {
        str << "mate " << -(Constants::MATE+score+1)/2;
    }
    else if (score >= Constants::MATE-256) {
        str << "mate " << (Constants::MATE-score+1)/2;
    }
    else
        str << "cp " << score;
}


int Scoring::tryBitbase(const Board &board)
{
    const Material &wMat = board.getMaterial(White);
    const Material &bMat = board.getMaterial(Black);
    int score = 0;
    if ((unsigned)wMat.infobits() == Material::KP &&
    bMat.kingOnly()) {
        Square passer = board.pawn_bits[White].firstOne();
        Square kp = board.kingSquare(White);
        if (lookupBitbase(board.kingSquare(White),passer,
            board.kingSquare(Black),White,board.sideToMove()))
            score += BITBASE_WIN;
        else
            return 0;
        int rank = Rank(passer,White);
        score += PASSED_PAWN[rank];
        int file_dist = Util::Abs(File(passer) - File(kp));
        int rank_dist = Rank(kp,White)-rank;
        // strongly encourage putting the king ahead of the
        // pawn
        if (rank_dist > 0 && file_dist <= 1)
            score += KING_AHEAD_OF_PAWN;
        else if (rank_dist<=0)                    // king behind pawn
            score += 8-rank*2;
        //also encourage staying near pawn
        score += DISTANCE_FROM_PAWN*distance1(kp,passer);
        return board.sideToMove() == White ? score : -score;
    }
    else if ((unsigned)bMat.infobits() == Material::KP &&
    wMat.kingOnly()) {
        Square passer = board.pawn_bits[Black].firstOne();
        Square kp = board.kingSquare(Black);
        if (lookupBitbase(board.kingSquare(Black),passer,
            board.kingSquare(White),Black,board.sideToMove()))
            score -= BITBASE_WIN;
        else
            return 0;
        int rank = Rank(passer,Black);
        score -= PASSED_PAWN[rank];
        int file_dist = Util::Abs(File(passer) - File(kp));
        int rank_dist = Rank(kp,Black)-rank;
        // strongly encourage putting the king ahead of the
        // pawn
        if (rank_dist > 0 && file_dist <= 1)
            score -= KING_AHEAD_OF_PAWN;
        else if (rank_dist<=0)                    // king behind pawn
            score -= 8-rank*2;
        //also encourage staying near pawn
        score -= DISTANCE_FROM_PAWN*distance1(kp,passer);
        return board.sideToMove() == White ? score : -score;
    }
    return INVALID_SCORE;
}

void Scoring::calcEndgame(const Board &board, PawnHashEntry *pawnEntry,
                          EndgameHashEntry *endgameEntry) {
    const PawnHashEntry::PawnData &wPawnData = pawnEntry->wPawnData;
    const PawnHashEntry::PawnData &bPawnData = pawnEntry->bPawnData;
    endgameEntry->wScore = endgameEntry->bScore = 0;
    int k_pos = 0; Square kp = board.kingSquare(White);
    // evaluate king position. Big bonus for centralizing king or (if
    // pawns are all on one side) being near pawns.
    // Similar to how Crafty does it.
    Bitboard all_pawns(board.pawn_bits[White] | board.pawn_bits[Black]);
    if (!all_pawns.is_clear()) {
       k_pos += KingEndgameScoresW[kp];
       if (!TEST_MASK(abcd_mask,all_pawns)) {
           if (File(kp)>DFILE)
             k_pos += PAWN_VALUE/8;
           else
             k_pos -= PAWN_VALUE/8;
       }
       else if (!TEST_MASK(efgh_mask,all_pawns)) {
           if (File(kp)<=DFILE)
             k_pos += PAWN_VALUE/8;
           else
             k_pos -= PAWN_VALUE/8;
       }
    }
    endgameEntry->white_king_position  = k_pos;
    k_pos = 0; kp = board.kingSquare(Black);
    if (!all_pawns.is_clear()) {
       k_pos += KingEndgameScoresB[kp];
       if (!TEST_MASK(abcd_mask,all_pawns)) {
           if (File(kp)>DFILE)
             k_pos += PAWN_VALUE/8;
           else
             k_pos -= PAWN_VALUE/8;
       }
       else if (!TEST_MASK(efgh_mask,all_pawns)) {
           if (File(kp)<=DFILE)
             k_pos += PAWN_VALUE/8;
           else
             k_pos -= PAWN_VALUE/8;
       }
    }
    endgameEntry->black_king_position = k_pos;
    endgameEntry->white_endgame_pawn_proximity =
        endgameEntry->black_endgame_pawn_proximity = (byte)0;
    endgameEntry->w_uncatchable = endgameEntry->b_uncatchable = (byte)0;
    if (!TEST_MASK(all_pawns,left_side_mask[4]) ||
        !TEST_MASK(all_pawns,right_side_mask[3])) {
        // bonus for king near pawns
        Bitboard it(all_pawns);
        Square sq;
        while (it.iterate(sq)) {
            endgameEntry->white_endgame_pawn_proximity += (4-distance1(board.kingSquare(White),sq));
            endgameEntry->black_endgame_pawn_proximity += (4-distance1(board.kingSquare(Black),sq));
        }
#ifdef EVAL_DEBUG
        cout << "king/pawn proximity (endgame) = " <<
            endgameEntry->white_endgame_pawn_proximity -
            endgameEntry->black_endgame_pawn_proximity << endl;
#endif
    }
    Bitboard passers2(wPawnData.passers);
    Square sq;
    while (passers2.iterate(sq)) {
      if (lookupBitbase(board.kingSquare(White),sq,
        board.kingSquare(Black),White,board.sideToMove())) {
#ifdef PAWN_DEBUG
           cout << "White pawn on " <<  SquareImage(sq) <<
                " is uncatchable" << endl;
#endif
           endgameEntry->w_uncatchable++;
        }
        int rank = Rank(sq,White);
        int file = File(sq);
        if (rank == 6 && (File(board.kingSquare(White)) == file-1 || File(board.kingSquare(White)) == file+1) &&
            Rank(board.kingSquare(White),White) >= rank)
            endgameEntry->wScore += SUPPORTED_PASSER6;
        else if (rank == 7 && (File(board.kingSquare(White)) == file-1 || File(board.kingSquare(White)) == file+1) &&
            Rank(board.kingSquare(White),White) >= rank)
            endgameEntry->wScore += SUPPORTED_PASSER7;
    }
    passers2 = bPawnData.passers;
    while (passers2.iterate(sq)) {
      if (lookupBitbase(board.kingSquare(Black),sq,
        board.kingSquare(White),Black,board.sideToMove())) {
#ifdef PAWN_DEBUG
           cout << "Black pawn on " << SquareImage(sq) << 
                   " is uncatchable" << endl;
#endif
           endgameEntry->b_uncatchable++;
        }
        int rank = Rank(sq,Black);
        int file = File(sq);
        if (rank == 6 && (File(board.kingSquare(Black)) == file-1 || File(board.kingSquare(Black)) == file+1) &&
            Rank(board.kingSquare(Black),Black) >= rank)
            endgameEntry->bScore += SUPPORTED_PASSER6;
        else if (rank == 7 && (File(board.kingSquare(Black)) == file-1 || File(board.kingSquare(Black)) == file+1) &&
            Rank(board.kingSquare(Black),Black) >= rank)
            endgameEntry->bScore += SUPPORTED_PASSER7;
    }
}

Move Scoring::getBestMove(const Board &board) const {
   EvalCacheEntry &entry = evalCache[board.hashCode() % EVAL_CACHE_SIZE];
   hash_t key = entry.move_key;
   Move best = entry.best;
   key ^= (hash_t)best;
   if (board.hashCode() == key) {
      return best;
   }
   else
      return NullMove;
}

void Scoring::cacheBestMove(const Board &board, Move best) {
   EvalCacheEntry &entry = evalCache[board.hashCode() % EVAL_CACHE_SIZE];
   entry.move_key = board.hashCode() ^ (hash_t)best;
   entry.best = best;
}

void Scoring::clearHashTables() {
    memset(pawnHashTable,0xff,PAWN_HASH_SIZE*sizeof(PawnHashEntry));
    memset(kingCoverHashTable,'\0',KING_COVER_HASH_SIZE*sizeof(KingCoverHashEntry));
    memset(endgameHashTable,'\0',ENDGAME_HASH_SIZE*sizeof(EndgameHashEntry));
    for (int i = 0; i < EVAL_CACHE_SIZE; i++) {
      evalCache[i].score_key = evalCache[i].move_key = (hash_t)0;
      evalCache[i].score = INVALID_SCORE;
      evalCache[i].best = NullMove;
    }
}
