// Copyright 1992-2008 by Jon Dart.  All Rights Reserved.

#ifndef _SCORING_H
#define _SCORING_H

#include "board.h"

#include <iostream>
using namespace std;

class Scoring
{
    // This class does static evaluation of chess positions.

    public:
		
    enum { INVALID_SCORE = -Constants::MATE };

    static void init();

    static void cleanup();

    Scoring();

    ~Scoring();
        
    // evaluate "board" from the perspective of the side to move.
    int evalu8( const Board &board );

    // evaluate "board" from the perspective of the side to move.
    int evalu8( const Board &board, int alpha, int beta );

    // return a positional score
    int positionalScore( const Board &board, 
			int alpha, int beta);

    // return a material score
    int materialScore( const Board &board );
		
    // checks for legal draws plus certain other theoretically
    // draw positions
    static int isDraw(const Board &board);
        
    // checks for draw by repetition (returning repetition count) + 
    // other draw situations as in isDraw above.
    static int isDraw( const Board &board, int &rep_count, int ply);

    static int repetitionDraw( const Board &board );

    static int materialDraw( const Board &board );

    // Check only for legal draws
    static int isLegalDraw(const Board &board);

    static int theoreticalDraw(const Board &board);

    // Turn a score into a formatted string (mate scores are
    // shown like +Mate6).
    static void printScore( int score, ostream & );

    // Output scores in the format required by the UCI protocol.
    static void printScoreUCI( int score, ostream & );

#ifdef EVAL_STATS
    static void clearStats();
#endif

    // Try to return a score based on bitbases, INVALID_SCORE if not found
    static int tryBitbase(const Board &board);

    void cacheBestMove(const Board &board, Move best);

    Move getBestMove(const Board &board) const;

    void clearHashTables();

 private:

    enum GamePhase { OPENING, MIDDLEGAME, ENDGAME_TRANSITION, ENDGAME };

    struct PawnHashEntry {

       hash_t hc;

       enum {PROTECTED_PP = 1, OUTSIDE_PP = 2};

       struct PawnData {
           Bitboard passers;
           byte weak_files;
           byte flags;
	   byte weakopen;
	   byte isolated_count;
           Bitboard opponent_pawn_attacks;
	   int first_passer,last_passer;
	   int passer_score;
  	   int center_score;
	   int score;
  	   int w_square_pawns,b_square_pawns;

	   int weakFile(int file) const {
	     return weak_files & (1<<(file-1));
	   }

           int outside() const {
             return flags & OUTSIDE_PP;
           }

           int protected_pp() const {
             return flags & PROTECTED_PP;
           }

       } wPawnData, bPawnData;

       char pad[8];

       const PawnData &pawnData(ColorType side) const {
	 return (side==White) ? wPawnData : bPawnData;
       }

    };

    PawnHashEntry *pawnHashTable;

    struct KingCoverHashEntry {
      hash_t hc;
      int cover;
      int pawn_proximity;
    };

    KingCoverHashEntry *kingCoverHashTable;

    struct EndgameHashEntry {
        // defines one entry in the king/pawn hash table.
        hash_t hc;                                    // hashcode
        signed char white_king_position, black_king_position;
        signed char white_endgame_pawn_proximity;
        signed char black_endgame_pawn_proximity;
        byte w_uncatchable, b_uncatchable;
        byte pawns_both_sides;
        int wScore,bScore;
    };

    EndgameHashEntry *endgameHashTable;


    int pieceScore(const Board &board, ColorType side, GamePhase phase,
                   const PawnHashEntry::PawnData &ourPawnData,
		   const PawnHashEntry::PawnData &oppPawnData,
		   int cover, int proximity);
    int calcPawnData(const Board &board, ColorType side,
			   PawnHashEntry::PawnData &entr);
    static int calcCover(const Board &board, ColorType side, Square kp);
    void calcCover(const Board &board, ColorType side,
		   KingCoverHashEntry &cover);
    void evalOutsidePassers(const Board &board,
			   PawnHashEntry *pawnEntry);
    void calcPawnEntry(const Board &board, PawnHashEntry *pawnEntry);

    int pawnScore(const Board &board, ColorType side,
		  const PawnHashEntry::PawnData &oppPawnData, GamePhase);

    void calcEndgame(const Board &board, PawnHashEntry *pawnEntry,
		     EndgameHashEntry *endgameEntry);

    int scoreEndgame(const Board &, EndgameHashEntry *endgameEntry,
		     const PawnHashEntry::PawnData &pawnData, ColorType side,
		     GamePhase phase);

    static GamePhase gamePhase(int level);
};

#endif



