// Copyright 2005-2010 by Jon Dart. All Rights Reserved.

#include "threadp.h"
#include "search.h"
#ifdef _THREAD_TRACE
#include "globals.h"
#endif
#ifndef _WIN32
#include <errno.h>
#include <fcntl.h>
#endif

#ifdef _THREAD_TRACE
static lock_t io_lock;

void log(const string &s) {
  Lock(io_lock);
  puts(s.c_str());
  fflush(stdout);
  Unlock(io_lock);
}
void log(const string &s,int param) {
  std::ostringstream out;
  out << s;
  out << ": ";
  out << param << '\0';
  log(out.str());
}
#endif

void idle_loop(ThreadInfo *ti, split_t *split) {

   while (ti->state != ThreadInfo::Terminating) {
#ifdef _THREAD_TRACE
      std::ostringstream s;
      s << "thread " << ti->index << " in idle loop:";
      s << "master=" << ti->master << " hasSplit=" << (split && split->slaves.size()) << '\0';
      log(s.str());
#endif
      ti->state = ThreadInfo::Idle; // mark thread available again
      int result;
      if ((result = ti->wait()) != 0) {
        if (result == -1) continue; // was interrupted
        else break;
      }
#ifdef _THREAD_TRACE
      log("unblocked",ti->index);
#endif
      // We've been woken up. There are three possible reasons:
      // 1. This thread is terminating.
      // 2. We are a master at a split point and all our slave threads
      // are done.
      // 3. This thread (a slave) has been assigned some work.
      //
      if (ti->state == ThreadInfo::Terminating) break;
#ifdef _THREAD_TRACE
      log("wakeup",ti->index);
#endif
      ASSERT(ti->work);
#ifdef _DEBUG
      int state = ti->state;
      ASSERT(state == ThreadInfo::Working);
#endif
      ti->work->searchSMP(ti);  
      ti->pool->checkIn(ti);
   }
}

#ifdef _WIN32
static DWORD WINAPI parkingLot(void *x)
#else
static void * CDECL parkingLot(void *x)
#endif
{
  idle_loop((ThreadInfo*)x,NULL);
  return 0;
}

ThreadInfo::~ThreadInfo() {
#ifdef _WIN32
  CloseHandle(hEvent1);
#elif defined(_MAC)
  if (semctl(sem,IPC_RMID,0)) perror("semctl");
#else
  if (sem_destroy(&sem)) perror("sem_destroy");
#endif
}

void ThreadInfo::start() {
    ASSERT(index>=0);
#ifdef _THREAD_TRACE
    log("start",index);
#endif
    signal();
}

void ThreadInfo::signal() {
   // unblock the thread so it can execute
   ASSERT(state != Idle);
#ifdef _WIN32
   if (!SetEvent(hEvent1)) {
      cerr << "error: SetEvent" << endl;
      ASSERT(0);
   }
#elif defined(_MAC)
    struct sembuf sb = { 0, -1, IPC_NOWAIT };
    if (semop(sem,&sb,1)&& errno!=EAGAIN) 
       perror("semop: signal");
#else // Linux
    if (sem_post(&sem) && errno!=EAGAIN) 
      perror("sem_post: signal");
#endif
}


void ThreadInfo::reset() {
#ifdef _THREAD_TRACE
  log("reset",index);
#endif
#ifdef _WIN32
   if (!ResetEvent(hEvent1)) {
        cerr << "ResetEvent failed" << endl;
        ASSERT(0);
   }
#elif defined(_MAC)
   // use semctl, not semop, to mimic Windows behavior where
   // multiple resets have no extra effect. 
   union semun val;
   // set semaphore to 1
   val.val = 1; 
   if (semctl(sem, 0, SETVAL, val)) {
      perror("semctl");
   }
#else
   // lock semaphore
   if (sem_trywait(&sem) && errno != EAGAIN) {
      perror("reset");
   }
#endif
}


int ThreadInfo::wait() {
#ifdef _THREAD_TRACE
   log("in wait",index);
#endif
#ifdef _WIN32
   return (WaitForSingleObject(hEvent1,INFINITE) != WAIT_OBJECT_0);
#elif defined(_MAC)
   struct sembuf sb = { 0, 0, 0 };
   // wait for semaphore value 0  
   if (semop(sem,&sb,1)) {
     if (errno == EINTR || errno == EIDRM) 
        return 1;
     else 
        perror("semop: wait");
   }
#else
   // wait for semaphore value 0  
   int ret;
   if ((ret = sem_wait(&sem)) != 0) {
      if (ret == EINTR) {
         return -1;
      }
      else if (ret == EIDRM) {
         return 1;
      }
      else {
         perror("semop: wait");
      }
   }
#endif
   return 0;
}

ThreadInfo::ThreadInfo(int i) 
 : state(Idle),
#ifdef _WIN32
   thread_id(NULL),
   work(NULL),
#else
   work(NULL),
#endif
   parentNode(NULL),
   master(0),
   index(i)
{
#ifdef _THREAD_TRACE
  log("starting",i);
#endif
#ifdef _WIN32
      // initial state is FALSE (non-signaled), so the
      // the controller blocks the thread
      hEvent1 = CreateEvent(NULL,FALSE,FALSE,NULL);
      DWORD id;
      if (index == 0) {
         thread_id = GetCurrentThread();
      } else {
         thread_id = CreateThread(NULL,0,
            parkingLot,this,
            0,
            &id);
      }
#else
#ifdef _MAC
      sem = semget(IPC_PRIVATE,1,IPC_CREAT | 0777);
      if (sem == -1) {
        perror("semaphore creation failed");
        return;
      }
      union semun val;
      // initialize semaphore to 1
      val.val = 1; 
      if (semctl(sem, 0, SETVAL, val)) {
        perror("semctl");
        return;
      }
#else
      // set val to zero so threads are blocked, initially
      if (sem_init(&sem,0,0)) {
         perror("semaphore creation failed");
         return;
      }
#endif
      if (index == 0) {
         thread_id = pthread_self();
      }
      else {
        if (pthread_create(&thread_id, NULL, parkingLot, this)) {
           perror("thread creation failed");
        }
      }
#endif
}

ThreadPool::ThreadPool(SearchController *controller,int n) 
: nthreads(n) {
#ifdef _THREAD_TRACE
  LockInit(io_lock);
#endif
   data = new ThreadInfo *[Constants::MaxCPUs];
   memset(data,'\0',sizeof(ThreadInfo*)*Constants::MaxCPUs);
   LockInit(poolLock);
   for (int i = 0; i < n; i++) {
      data[i] = new ThreadInfo(i);
      data[i]->pool = this;
      if (i==0) {
         data[i]->work = new RootSearch(controller,data[i]);
      }
      else {
         data[i]->work = new Search(controller,data[i]);
      }
      data[i]->work->ti = data[i];
      data[i]->pool = this;
   }
}

ThreadPool::~ThreadPool() {
    shutDown();
#ifdef _THREAD_TRACE
  LockDestroy(io_lock);
#endif
}

void ThreadPool::shutDown() {
    // do not shut down thread 0 (main thread)
    for (int i = 1; i < nthreads; i++) {
       // All threads should be idle when this function is called.
       if (data[i]->state == ThreadInfo::Idle) {
          // Set the thread to the terminating state that will force thread
          // procedure exit
          data[i]->state = ThreadInfo::Terminating;
          // unblock the thread
          data[i]->signal();
       }
       // wait for the thread to terminate
#ifdef _WIN32
       WaitForSingleObject(data[i]->thread_id,INFINITE);
#else
       void *value_ptr;
       pthread_join(data[i]->thread_id,&value_ptr);
#endif
       // Free thread data 
       delete data[i];
    }
    for (int i = nthreads; i < Constants::MaxCPUs; i++) {
      // Free thread data - may be non-zero if we have shrunk the thread pool
      delete data[i];
    }
    delete data[0]; // main thread structure
}

ThreadInfo * ThreadPool::checkOut(Search *parent, NodeInfo *forNode,
  int ply, int depth) {
    Lock(poolLock);
    for (int i = 0; i < nthreads; i++) {
#ifdef _WIN32
       if (i!=parent->ti->index && data[i]->state == ThreadInfo::Idle) {
#elif defined(_MAC)
       if (i!=parent->ti->index && data[i]->state == ThreadInfo::Idle &&
           semctl(data[i]->sem,0,GETZCNT)>0) {
#else // Linux
         // verify the thread is available and is actually
         // at the synchronization point
         int val = 0;
         if (sem_getvalue(&(data[i]->sem),&val)) {
           perror("sem_get");
           continue;
         }
         if (i!=parent->ti->index && data[i]->state == ThreadInfo::Idle &&
            val==0) {
#endif
         // if this is a "master" thread it is not sufficient to just be
         // idle - assign it only to one of its slave threads
         if (!data[i]->master || 
             (data[i]->parentNode->split &&
             data[i]->parentNode->split->slaves.exists(parent->ti))) {
           split_t *split = parent->split(forNode,data[i],ply,depth);
           if (!split) {
              Unlock(poolLock);
#ifdef _THREAD_TRACE
              log("alloc failed, split limit");
#endif
              return NULL; // can't split this search instance any more
           }
           // We're working now - ensure we will not be allocated again
           data[i]->state = ThreadInfo::Working; 
           Unlock(poolLock);
           return data[i]; 
         }
       }
    }
    // no luck, no free threads
    Unlock(poolLock);
    return NULL;
}

void ThreadPool::resize(int n, SearchController *controller) {
  if (n >= 1 && n < Constants::MaxCPUs && n != nthreads) {
    Lock(poolLock);
    if (n>nthreads) {
      data[nthreads] = new ThreadInfo(nthreads);
      data[nthreads]->work = new Search(controller,data[nthreads]);
      data[nthreads]->pool = this;
      nthreads++;
    }
    else {
      // shrinking
      --nthreads;
      data[nthreads]->state = ThreadInfo::Terminating;
      if (data[nthreads-1]->state == ThreadInfo::Idle) {
        data[nthreads]->signal(); // unblock thread & exit thread proc
      }
    }
    Unlock(poolLock);
  }
}

void ThreadPool::checkIn(ThreadInfo *ti) {
#ifdef _THREAD_TRACE
  std::ostringstream s;
  s << "checkIn " << ti->index << " master=" << ti->master << '\0';
  log(s.str());
#endif
    Lock(poolLock);
    if (!ti->master) {
      Search *parent = ti->work->parent;
      if (!parent->join(ti->parentNode,ti)) {
        // all slave threads are completed, so signal parent that it
        // does not need to wait any more
#ifdef _THREAD_TRACE
        std::ostringstream s;
        s << "signaling parent (" << parent->ti->index << ")" << 
          " parent state=" << parent->ti->state << '\0';
        log(s.str());
#endif
        ASSERT(parent->ti->index != ti->index);
        parent->ti->signal();
      }
      // semaphores are not auto-reset, so increment the value
      // here so we will wait when we re-enter the idle loop.
      ti->reset();
#if defined(_WIN32) && defined(_DEBUG)
      // this event should be non-signaled (so we will wait on 
      // next entry into idle loop).
      int result = WaitForSingleObject(ti->hEvent1,0);
      ASSERT(WAIT_TIMEOUT == result);
#endif
   }
   Unlock(poolLock);
}

int ThreadPool::activeCount() {
    int count = 0;
    for (int i = 0; i < nthreads; i++) {
       if (data[i]->state == ThreadInfo::Working) ++count;
    }
    return count;
}

void ThreadPool::clearHashTables() {
  for (int i=0;i< nthreads;i++) {
    if (data[i]->work) data[i]->work->clearHashTables();
  }
}
